__all__ = [
    "order_series_by",
    "desc",
    "coalesce",
    "case_when",
    "if_else",
    "na_if",
]

import collections
from .base import make_symbolic
from numpy import argmin, arange, unique, repeat, nan, array
from pandas import concat, DataFrame, Series, isnull


# ------------------------------------------------------------------------------
# series ordering
# ------------------------------------------------------------------------------


@make_symbolic
def order_series_by(series, order_series):
    """
    Orders one series according to another series, or a list of other
    series. If a list of other series are specified, ordering is done hierarchically
    like when a list of columns is supplied to `.sort_values()`.

    Args:
        series (:obj:`pandas.Series`): the pandas Series object to be reordered.
        order_series: either a pandas Series object or a list of pandas Series
            objects. These will be sorted using `.sort_values()` with
            `ascending=True`, and the new order will be used to reorder the
            Series supplied in the first argument.

    Returns:
        reordered `pandas.Series` object
    """

    if isinstance(order_series, (list, tuple)):
        sorter = concat(order_series, axis=1)
        sorter_columns = ["_sorter" + str(i) for i in range(len(order_series))]
        sorter.columns = sorter_columns
        sorter["series"] = series.values
        sorted_series = sorter.sort_values(sorter_columns)["series"]
        return sorted_series

    sorted_series = DataFrame(
        {"series": series.values, "order": order_series.values}
    ).sort_values("order", ascending=True)["series"]
    return sorted_series


@make_symbolic
def desc(series):
    """
    Mimics the functionality of the R desc function. Essentially inverts a
    series object to make ascending sort act like descending sort.

    Args:
        series (:obj:`pandas.Series`): pandas series to be inverted prior to
            ordering/sorting.

    Returns:
        inverted `pandas.Series`. The returned series will be numeric (integers),
            regardless of the type of the original series.

    Example:

        First group by cut, then find the first value of price when ordering by
        price ascending, and ordering by price descending using the `desc` function.

        diamonds >> group_by(X.cut) >> summarize(carat_low=first(X.price, order_by=X.price),
                                                 carat_high=first(X.price, order_by=desc(X.price)))

                 cut  carat_high  carat_low
        0       Fair       18574        337
        1       Good       18788        327
        2      Ideal       18806        326
        3    Premium       18823        326
        4  Very Good       18818        336
    """

    return series.rank(method="min", ascending=False)


# ------------------------------------------------------------------------------
# coalesce
# ------------------------------------------------------------------------------


@make_symbolic
def coalesce(*series):
    """
    Takes the first non-NaN value in order across the specified series,
    returning a new series. Mimics the coalesce function in dplyr and SQL.

    Args:
        *series: Series objects, typically represented in their symbolic form
            (like X.series).

    Example:
        df = pd.DataFrame({
            'a':[1,np.nan,np.nan,np.nan,np.nan],
            'b':[2,3,np.nan,np.nan,np.nan],
            'c':[np.nan,np.nan,4,5,np.nan],
            'd':[6,7,8,9,np.nan]
        })
        df >> transmute(coal=coalesce(X.a, X.b, X.c, X.d))

             coal
        0       1
        1       3
        2       4
        3       5
        4  np.nan
    """

    series = [Series(s) for s in series]
    coalescer = concat(series, axis=1)
    min_nonna = argmin(isnull(coalescer).values, axis=1)
    min_nonna = [coalescer.columns[i] for i in min_nonna]
    return coalescer.lookup(arange(coalescer.shape[0]), min_nonna)


# ------------------------------------------------------------------------------
# case_when
# ------------------------------------------------------------------------------


@make_symbolic
def case_when(*conditions):
    """
    Functions as a switch statement, creating a new series out of logical
    conditions specified by 2-item lists where the left-hand item is the
    logical condition and the right-hand item is the value where that
    condition is true.

    Conditions should go from the most specific to the most general. A
    conditional that appears earlier in the series will "overwrite" one that
    appears later. Think of it like a series of if-else statements.

    The logicals and values of the condition pairs must be all the same
    length, or length 1. Logicals can be vectors of booleans or a single
    boolean (`True`, for example, can be the logical statement for the
    final conditional to catch all remaining.).

    Args:
        *conditions: Each condition should be a list with two values. The first
            value is a boolean or vector of booleans that specify indices in
            which the condition is met. The second value is a vector of values
            or single value specifying the outcome where that condition is met.

    Example:
        df = pd.DataFrame({
            'num':np.arange(16)
        })
        df >> mutate(strnum=case_when([X.num % 15 == 0, 'fizzbuzz'],
                                      [X.num % 3 == 0, 'fizz'],
                                      [X.num % 5 == 0, 'buzz'],
                                      [True, X.num.astype(str)]))

            num    strnum
        0     0  fizzbuzz
        1     1         1
        2     2         2
        3     3      fizz
        4     4         4
        5     5      buzz
        6     6      fizz
        7     7         7
        8     8         8
        9     9      fizz
        10   10      buzz
        11   11        11
        12   12      fizz
        13   13        13
        14   14        14
        15   15  fizzbuzz
    """

    lengths = []
    for logical, outcome in conditions:
        if isinstance(logical, collections.Iterable):
            lengths.append(len(logical))
        if isinstance(outcome, collections.Iterable) and not isinstance(outcome, str):
            lengths.append(len(outcome))
    unique_lengths = unique(lengths)
    assert len(unique_lengths) == 1
    output_len = unique_lengths[0]

    output = []
    for logical, outcome in conditions:
        if isinstance(logical, bool):
            logical = repeat(logical, output_len)
        if isinstance(logical, Series):
            logical = logical.values
        if not isinstance(outcome, collections.Iterable) or isinstance(outcome, str):
            outcome = Series(repeat(outcome, output_len))
        outcome[~logical] = nan
        output.append(outcome)

    return coalesce(*output)


# ------------------------------------------------------------------------------
# if_else
# ------------------------------------------------------------------------------


@make_symbolic
def if_else(condition, when_true, otherwise):
    """
    Wraps creation of a series based on if-else conditional logic into a function
    call.

    Provide a boolean vector condition, value(s) when true, and value(s)
    when false, and a vector will be returned the same length as the conditional
    vector according to the logical statement.

    Args:
        condition: A boolean vector representing the condition. This is often
            a logical statement with a symbolic series.
        when_true: A vector the same length as the condition vector or a single
            value to apply when the condition is `True`.
        otherwise: A vector the same length as the condition vector or a single
            value to apply when the condition is `False`.

    Example:
    df = pd.DataFrame
    """

    if not isinstance(when_true, collections.Iterable) or isinstance(when_true, str):
        when_true = repeat(when_true, len(condition))
    if not isinstance(otherwise, collections.Iterable) or isinstance(otherwise, str):
        otherwise = repeat(otherwise, len(condition))
    assert (len(condition) == len(when_true)) and (len(condition) == len(otherwise))

    if isinstance(when_true, Series):
        when_true = when_true.values
    if isinstance(otherwise, Series):
        otherwise = otherwise.values

    output = array(
        [when_true[i] if c else otherwise[i] for i, c in enumerate(condition)]
    )
    return output


# ------------------------------------------------------------------------------
# na_if
# ------------------------------------------------------------------------------


@make_symbolic
def na_if(series, *values):
    """
    If values in a series match a specified value, change them to `np.nan`.

    Args:
        series: Series or vector, often symbolic.
        *values: Value(s) to convert to `np.nan` in the series.
    """

    series = Series(series)
    series[series.isin(values)] = nan
    return series



__all__ = [
    "eval_contour",
    "ev_contour",
]

from grama import eval_df, add_pipe, tf_outer
from numpy import array, linspace, isfinite, reshape, full
from pandas import concat, DataFrame
from toolz import curry

class Square():
    # D -- C
    # |    |
    # |    |
    # A -- B
    A = [0, 0]
    B = [0, 0]
    C = [0, 0]
    D = [0, 0]
    A_data = 0.0
    B_data = 0.0
    C_data = 0.0
    D_data = 0.0

    def GetCaseId(self, threshold):
        caseId = 0
        if (self.A_data >= threshold):
            caseId |= 1
        if (self.B_data >= threshold):
            caseId |= 2
        if (self.C_data >= threshold):
            caseId |= 4
        if (self.D_data >= threshold):
            caseId |= 8

        return caseId

    def InterpolateVertical(self, px, py, qx, qy, d, d1, thresh):
        x = px
        y = (
            py +
            (qy - py) * ((thresh - d) / (d1 - d))
        )
        return x, y

    def InterpolateHorizontal(self, px, py, qx, qy, d, d1, thresh):
        y = py
        x = (
            px +
            (qx - px) * ((thresh - d) / (d1 - d))
        )
        return x, y

    def GetLines(self, Threshold):
        lines = []
        caseId = self.GetCaseId(Threshold)

        if caseId in (0, 15):
            return []

        if caseId in (1, 14, 10):
            pX, pY = self.InterpolateVertical(
                *self.A, *self.D, self.A_data, self.D_data, Threshold
            )
            qX, qY = self.InterpolateHorizontal(
                *self.A, *self.B, self.A_data, self.B_data, Threshold
            )

            line = (pX, pY, qX, qY)

            lines.append(line)

        if caseId in (2, 13, 5):
            pX, pY = self.InterpolateHorizontal(
                *self.A, *self.B, self.A_data, self.B_data, Threshold
            )
            qX, qY = self.InterpolateVertical(
                *self.B, *self.C, self.B_data, self.C_data, Threshold
            )

            line = (pX, pY, qX, qY)

            lines.append(line)

        if caseId in (3, 12):
            pX, pY = self.InterpolateVertical(
                *self.A, *self.D, self.A_data, self.D_data, Threshold
            )
            qX, qY = self.InterpolateVertical(
                *self.C, *self.B, self.C_data, self.B_data, Threshold
            )

            line = (pX, pY, qX, qY)

            lines.append(line)

        if caseId in (4, 11, 10):
            pX, pY = self.InterpolateHorizontal(
                *self.D, *self.C, self.D_data, self.C_data, Threshold
            )
            qX, qY = self.InterpolateVertical(
                *self.B, *self.C, self.B_data, self.C_data, Threshold
            )

            line = (pX, pY, qX, qY)

            lines.append(line)

        elif caseId in (6, 9):
            pX, pY = self.InterpolateHorizontal(
                *self.D, *self.C, self.D_data, self.C_data, Threshold
            )
            qX, qY = self.InterpolateHorizontal(
                *self.A, *self.B, self.A_data, self.B_data, Threshold
            )

            line = (pX, pY, qX, qY)

            lines.append(line)

        elif caseId in (7, 8, 5):
            pX, pY = self.InterpolateVertical(
                *self.A, *self.D, self.A_data, self.D_data, Threshold
            )
            qX, qY = self.InterpolateHorizontal(
                *self.D, *self.C, self.D_data, self.C_data, Threshold
            )

            line = (pX, pY, qX, qY)

            lines.append(line)

        return lines

def marching_square(xVector, yVector, Data, threshold):
    linesList = []

    Height = len(Data)  # rows
    Width = len(Data[1])  # cols

    if ((Width == len(xVector)) and (Height == len(yVector))):
        squares = full((Height - 1, Width - 1), Square())

        sqHeight = squares.shape[0]  # rows count
        sqWidth = squares.shape[1]  # cols count

        for j in range(sqHeight):  # rows
            for i in range(sqWidth):  # cols
                a = Data[j + 1, i]
                b = Data[j + 1, i + 1]
                c = Data[j, i + 1]
                d = Data[j, i]
                A = [xVector[i], yVector[j + 1]]
                B = [xVector[i + 1], yVector[j + 1]]
                C = [xVector[i + 1], yVector[j]]
                D = [xVector[i], yVector[j]]

                squares[j, i].A_data = a
                squares[j, i].B_data = b
                squares[j, i].C_data = c
                squares[j, i].D_data = d

                squares[j, i].A = A
                squares[j, i].B = B
                squares[j, i].C = C
                squares[j, i].D = D

                listTemp = squares[j, i].GetLines(threshold)

                linesList = linesList + listTemp
    else:
        raise AssertionError

    return [linesList]

## Generate contours from a model
# --------------------------------------------------
@curry
def eval_contour(
        model,
        var=None,
        out=None,
        df=None,
        levels=None,
        n_side=128,
        n_levels=5,
):
    r"""Generate contours from a model

    Generates contours from a model. Evaluates the model on a dense grid, then
    runs marching squares to generate contours. Supports targeting multiple
    outputs and handling auxiliary inputs not included in the contour map.

    Args:
        model (gr.Model): Model to evaluate.
        var (list of str): Model inputs to target; must provide exactly
            two inputs, and both must have finite domain width.
        out (list of str): Model output(s) for contour generation.
        df (DataFrame): Levels for model variables not included in var (auxiliary inputs).
        levels (dict): Specific output levels for contour generation;
            overrides n_levels.
        n_side (int): Side resolution for grid; n_side**2 total evaluations.
        n_levels (int): Number of contour levels.

    Returns:
        DataFrame: Points along contours, organized by output and auxiliary variable levels.

    Examples:

        >>> import grama as gr
        >>> ## Multiple outputs
        >>> (
        >>>     gr.Model()
        >>>     >> gr.cp_vec_function(
        >>>         fun=lambda df: gr.df_make(
        >>>             f=df.x**2 + df.y**2,
        >>>             g=df.x + df.y,
        >>>         ),
        >>>         var=["x", "y"],
        >>>         out=["f", "g"],
        >>>     )
        >>>     >> gr.cp_bounds(
        >>>         x=(-1, +1),
        >>>         y=(-1, +1),
        >>>     )
        >>>     >> gr.ev_contour(
        >>>         var=["x", "y"],
        >>>         out=["f", "g"],
        >>>     )
        >>>
        >>>     >> gr.ggplot(gr.aes("x", "y"))
        >>>     + gr.geom_segment(gr.aes(xend="x_end", yend="y_end", group="level", color="out"))
        >>> )
        >>> ## Auxiliary inputs
        >>> (
        >>>     gr.Model()
        >>>     >> gr.cp_vec_function(
        >>>         fun=lambda df: gr.df_make(
        >>>             f=df.c * df.x + (1 - df.c) * df.y,
        >>>         ),
        >>>         var=["x", "y"],
        >>>         out=["f", "g"],
        >>>     )
        >>>     >> gr.cp_bounds(
        >>>         x=(-1, +1),
        >>>         y=(-1, +1),
        >>>     )
        >>>     >> gr.ev_contour(
        >>>         var=["x", "y"],
        >>>         out=["f"],
        >>>         df=gr.df_make(c=[0, 1])
        >>>     )
        >>>
        >>>     >> gr.ggplot(gr.aes("x", "y"))
        >>>     + gr.geom_segment(gr.aes(xend="x_end", yend="y_end", group="level", color="c"))
        >>> )

    """
    ## Check invariants
    # Argument given
    if var is None:
        raise ValueError("No `var` given")
    # Correct number of inputs
    if len(var) != 2:
        raise ValueError("Must provide exactly 2 inputs in `var`.")
    # Inputs available
    var_diff = set(var).difference(set(model.var))
    if len(var_diff) > 0:
        raise ValueError(
            "`var` must be a subset of model.var; missing: {}".format(var_diff)
        )
    # All inputs supported
    var_diff = set(model.var).difference(set(var))
    if len(var_diff) > 0:
        if df is None:
            raise ValueError(
                "Must provide values for remaining model variables using df; " +
                "missing values: {}".format(var_diff)
            )
        var_diff2 = var_diff.difference(set(df.columns))
        if len(var_diff2) > 0:
            raise ValueError(
                "All model variables need values in provided df; " +
                "missing values: {}".format(var_diff2)
            )
    # Finite bound width
    if not all([
            isfinite(model.domain.get_width(v)) and
            (model.domain.get_width(v) > 0)
            for v in var
    ]):
        raise ValueError("All model bounds for `var` must be finite and nonzero")

    # Argument given
    if out is None:
        raise ValueError("No `out` given")
    # Outputs available
    out_diff = set(out).difference(set(model.out))
    if len(out_diff) > 0:
        raise ValueError(
            "`out` must be a subset of model.out; missing: {}".format(out_diff)
        )

    ## Generate data
    xv = linspace(*model.domain.get_bound(var[0]), n_side)
    yv = linspace(*model.domain.get_bound(var[1]), n_side)
    df_x = DataFrame({var[0]: xv})
    df_y = DataFrame({var[1]: yv})
    df_input = (
        df_x
        >> tf_outer(df_outer=df_y)
    )

    # Create singleton level if necessary
    if df is None:
        df = DataFrame({"_foo":[0]})

    ## Loop over provided auxiliary levels
    df_res = DataFrame()
    for i in range(df.shape[0]):
        df_in_tmp = (
            df_input
            >> tf_outer(df_outer=df.iloc[[i]])
        )
        df_out = eval_df(
            model,
            df=df_in_tmp,
        )

        ## Set output threshold levels
        if levels is None:
            levels = dict(zip(
                out,
                [
                    linspace(df_out[o].min(), df_out[o].max(), n_levels + 2)[1:-1]
                    for o in out
                ]
            ))

        ## Run marching squares
        # Output quantity
        for o in out:
            # Reshape data
            Data = reshape(df_out[o].values, (n_side, n_side))
            # Threshold level
            for t in levels[o]:
                # Run marching squares
                segments = marching_square(xv, yv, Data, t)

                # Package
                df_tmp = DataFrame(
                    data=array(segments).squeeze(),
                    columns=[var[0], var[1], var[0]+"_end", var[1]+"_end"],
                )
                df_tmp["out"] = [o] * df_tmp.shape[0]
                df_tmp["level"] = [t] * df_tmp.shape[0]
                df_tmp = (
                    df_tmp
                    >> tf_outer(df_outer=df.iloc[[i]])
                )

                df_res = concat((df_res, df_tmp), axis=0)

    ## Remove dummy column, if present
    if "_foo" in df_res.columns:
        df_res.drop("_foo", axis=1, inplace=True)

    ## Return the results
    return df_res.reset_index(drop=True)


ev_contour = add_pipe(eval_contour)



__all__ = [
    "fit_gp",
    "ft_gp",
    "fit_lm",
    "ft_lm",
    "fit_rf",
    "ft_rf",
    "fit_kmeans",
    "ft_kmeans",
]

## Fitting via sklearn package
try:
    from sklearn.linear_model import LinearRegression
    from sklearn.gaussian_process import GaussianProcessRegressor
    from sklearn.gaussian_process.kernels import Kernel, RBF, ConstantKernel as Con
    from sklearn.cluster import KMeans
    from sklearn.ensemble import RandomForestRegressor

except ModuleNotFoundError:
    pass

from copy import deepcopy
from grama import add_pipe, cp_vec_function, Function, Model, pipe
from pandas import concat, DataFrame, Series
from toolz import curry
from warnings import filterwarnings


## Helper functions and classes
# --------------------------------------------------
def standardize_cols(df, ser_min, ser_max, var):
    """
    @pre set(ser_min.index) == set(ser_max.index)
    """

    df_std = df.copy()
    for v in var:
        den = ser_max[v] - ser_min[v]
        if den < 1e-16:
            den = 1
        df_std[v] = (df_std[v] - ser_min[v]) / den
    return df_std


def restore_cols(df, ser_min, ser_max, var):
    """
    @pre set(ser_min.index) == set(ser_max.index)
    """
    df_res = df.copy()
    for v in var:
        den = ser_max[v] - ser_min[v]
        if den < 1e-16:
            den = 1
        df_res[v] = den * df[v] + ser_min[v]
    return df_res


class FunctionGPR(Function):
    def __init__(self, gpr, var, out, name, runtime, var_min, var_max):
        self.gpr = gpr
        # self.df_train = df_train
        self.var = var
        ## "Natural" outputs; what we're modeling
        self.out_nat = out
        ## Predicted outputs; mean and std
        self.out_mean = list(map(lambda s: s + "_mean", out))
        self.out_sd = list(map(lambda s: s + "_sd", out))
        self.out = self.out_mean + self.out_sd

        self.name = name
        self.runtime = runtime

        self.var_min = var_min
        self.var_max = var_max

    def eval(self, df):
        ## Check invariant; model inputs must be subset of df columns
        if not set(self.var).issubset(set(df.columns)):
            raise ValueError(
                "Model function `{}` var not a subset of given columns".format(
                    self.name
                )
            )
        df_sd = standardize_cols(df, self.var_min, self.var_max, self.var)
        y, y_sd = self.gpr.predict(df_sd[self.var], return_std=True)

        return concat(
            (
                DataFrame(data=y, columns=self.out_mean),
                DataFrame(data=y_sd, columns=self.out_sd),
            ),
            axis=1,
        )

    def copy(self):
        func_new = FunctionGPR(
            self.gpr,
            self.df_train.copy(),
            self.var,
            self.out_nat,
            self.name,
            self.runtime,
        )

        return func_new


class FunctionRegressor(Function):
    def __init__(self, regressor, var, out, name, runtime):
        """

        Args:
            regressor (scikit Regressor):
        """
        self.regressor = regressor
        self.var = var
        self.out = list(map(lambda s: s + "_mean", out))
        self.name = name
        self.runtime = runtime

    def eval(self, df):
        ## Check invariant; model inputs must be subset of df columns
        if not set(self.var).issubset(set(df.columns)):
            raise ValueError(
                "Model function `{}` var not a subset of given columns".format(
                    self.name
                )
            )

        ## Predict
        y = self.regressor.predict(df[self.var])
        return DataFrame(data=y, columns=self.out)


## Fit GP model with sklearn
# --------------------------------------------------
@curry
def fit_gp(
    df,
    md=None,
    var=None,
    out=None,
    domain=None,
    density=None,
    kernels=None,
    seed=None,
    suppress_warnings=True,
    n_restart=5,
    alpha=1e-10,
):
    r"""Fit a gaussian process

    Fit a gaussian process to given data. Specify var and out, or inherit from
    an existing model.

    Note that the new model will have two outputs `y_mean, y_sd` for each
    original output `y`. The quantity `y_mean` is the best-fit value, while
    `y_sd` is a measure of predictive uncertainty.

    Args:
        df (DataFrame): Data for function fitting
        md (gr.Model): Model from which to inherit metadata
        var (list(str) or None): List of features or None for all except outputs
        out (list(str)): List of outputs to fit
        domain (gr.Domain): Domain for new model
        density (gr.Density): Density for new model
        seed (int or None): Random seed for fitting process
        kernels (sklearn.gaussian_process.kernels.Kernel or dict or None): Kernel for GP
        n_restart (int): Restarts for optimization
        alpha (float or iterable): Value added to diagonal of kernel matrix
        suppress_warnings (bool): Suppress warnings when fitting?

    Returns:
        gr.Model: A grama model with fitted function(s)

    Notes:
        - Wrapper for sklearn.gaussian_process.GaussianProcessRegressor

    """
    if suppress_warnings:
        filterwarnings("ignore")

    n_obs, n_in = df.shape

    ## Infer fitting metadata, if available
    if not (md is None):
        domain = md.domain
        density = md.density
        out = md.out

    ## Check invariants
    if not set(out).issubset(set(df.columns)):
        raise ValueError("out must be subset of df.columns")
    ## Default input value
    if var is None:
        var = list(set(df.columns).difference(set(out)))
    ## Check more invariants
    set_inter = set(out).intersection(set(var))
    if len(set_inter) > 0:
        raise ValueError(
            "out and var must be disjoint; intersect = {}".format(set_inter)
        )
    if not set(var).issubset(set(df.columns)):
        raise ValueError("var must be subset of df.columns")

    ## Pre-process kernel selection
    try:
        if kernels is None:
            # Vectorize
            kernels = {o: None for o in out}
        elif isinstance(kernels, Kernel):
            kernels = {o: kernels for o in out}

    except NameError as e:
        error_string = str(e)
        raise NameError(
            error_string +
            "\n\nThis function requires the `sklearn` package. " +
            "Try running the following to install the package:\n"
            "    pip install scikit-learn"
        )

    ## Pre-process data
    var_min = df[var].min()
    var_max = df[var].max()
    df_sd = standardize_cols(df, var_min, var_max, var)

    ## Construct gaussian process for each output
    functions = []

    try:
        for output in out:
            # Define and fit model
            gpr = GaussianProcessRegressor(
                kernel=deepcopy(kernels[output]),
                random_state=seed,
                normalize_y=True,
                copy_X_train=True,
                n_restarts_optimizer=n_restart,
                alpha=alpha,
            )
            gpr.fit(df_sd[var], df_sd[output])
            name = "GP ({})".format(str(gpr.kernel_))

            fun = FunctionGPR(gpr, var, [output], name, 0, var_min, var_max)
            functions.append(fun)

    except NameError as e:
        error_string = str(e)
        raise NameError(
            error_string +
            "\n\nThis function requires the `sklearn` package. " +
            "Try running the following to install the package:\n"
            "    pip install scikit-learn"
        )

    ## Construct model
    return Model(functions=functions, domain=domain, density=density)


ft_gp = add_pipe(fit_gp)

## Fit random forest model with sklearn
# --------------------------------------------------
@curry
def fit_rf(
    df,
    md=None,
    var=None,
    out=None,
    domain=None,
    density=None,
    seed=None,
    suppress_warnings=True,
    **kwargs
):
    r"""Fit a random forest

    Fit a random forest to given data. Specify inputs and outputs, or inherit
    from an existing model.

    Args:
        df (DataFrame): Data for function fitting
        md (gr.Model): Model from which to inherit metadata
        var (list(str) or None): List of features or None for all except outputs
        out (list(str)): List of outputs to fit
        domain (gr.Domain): Domain for new model
        density (gr.Density): Density for new model
        seed (int or None): Random seed for fitting process
        suppress_warnings (bool): Suppress warnings when fitting?

    Keyword Arguments:
        n_estimators (int):
        criterion (int):
        max_depth (int or None):
        min_samples_split (int, float):
        min_samples_leaf (int, float):
        min_weight_fraction_leaf (float):
        max_features (int, float, string):
        max_leaf_nodes (int or None):
        min_impurity_decrease (float):
        min_impurity_split (float):
        bootstrap (bool):
        oob_score (bool):
        n_jobs (int or None):
        random_state (int):

    Returns:
        gr.Model: A grama model with fitted function(s)

    Notes:
        - Wrapper for sklearn.ensemble.RandomForestRegressor

    """
    if suppress_warnings:
        filterwarnings("ignore")

    n_obs, n_in = df.shape

    ## Infer fitting metadata, if available
    if not (md is None):
        domain = md.domain
        density = md.density
        out = md.out

    ## Check invariants
    if not set(out).issubset(set(df.columns)):
        raise ValueError("out must be subset of df.columns")
    ## Default input value
    if var is None:
        var = list(set(df.columns).difference(set(out)))
    ## Check more invariants
    set_inter = set(out).intersection(set(var))
    if len(set_inter) > 0:
        raise ValueError(
            "outputs and inputs must be disjoint; intersect = {}".format(set_inter)
        )
    if not set(var).issubset(set(df.columns)):
        raise ValueError("var must be subset of df.columns")

    ## Construct gaussian process for each output
    functions = []

    try:
        for output in out:
            rf = RandomForestRegressor(random_state=seed, **kwargs)
            rf.fit(df[var], df[output])
            name = "RF"

            fun = FunctionRegressor(rf, var, [output], name, 0)
            functions.append(fun)

    except NameError as e:
        error_string = str(e)
        raise NameError(
            error_string +
            "\n\nThis function requires the `sklearn` package. " +
            "Try running the following to install the package:\n"
            "    pip install scikit-learn"
        )

    ## Construct model
    return Model(functions=functions, domain=domain, density=density)


ft_rf = add_pipe(fit_rf)

## Fit linear model with sklearn
# --------------------------------------------------
@curry
def fit_lm(
    df,
    md=None,
    var=None,
    out=None,
    domain=None,
    density=None,
    seed=None,
    suppress_warnings=True,
    **kwargs
):
    r"""Fit a linear model

    Fit a linear model to given data. Specify inputs and outputs, or inherit
    from an existing model.

    Args:
        df (DataFrame): Data for function fitting
        md (gr.Model): Model from which to inherit metadata
        var (list(str) or None): List of features or None for all except outputs
        out (list(str)): List of outputs to fit
        domain (gr.Domain): Domain for new model
        density (gr.Density): Density for new model
        seed (int or None): Random seed for fitting process
        suppress_warnings (bool): Suppress warnings when fitting?

    Returns:
        gr.Model: A grama model with fitted function(s)

    Notes:
        - Wrapper for sklearn.ensemble.RandomForestRegressor

    """
    if suppress_warnings:
        filterwarnings("ignore")

    n_obs, n_in = df.shape

    ## Infer fitting metadata, if available
    if not (md is None):
        domain = md.domain
        density = md.density
        out = md.out

    ## Check invariants
    if not set(out).issubset(set(df.columns)):
        raise ValueError("out must be subset of df.columns")
    ## Default input value
    if var is None:
        var = list(set(df.columns).difference(set(out)))
    ## Check more invariants
    set_inter = set(out).intersection(set(var))
    if len(set_inter) > 0:
        raise ValueError(
            "outputs and inputs must be disjoint; intersect = {}".format(set_inter)
        )
    if not set(var).issubset(set(df.columns)):
        raise ValueError("var must be subset of df.columns")

    ## Construct gaussian process for each output
    functions = []

    try:
        for output in out:
            lm = LinearRegression(**kwargs)
            lm.fit(df[var], df[output])
            name = "LM"

            fun = FunctionRegressor(lm, var, [output], name, 0)
            functions.append(fun)

    except NameError as e:
        error_string = str(e)
        raise NameError(
            error_string +
            "\n\nThis function requires the `sklearn` package. " +
            "Try running the following to install the package:\n"
            "    pip install scikit-learn"
        )

    ## Construct model
    return Model(functions=functions, domain=domain, density=density)


ft_lm = add_pipe(fit_lm)


## Fit kmeans clustering model
# --------------------------------------------------
@curry
def fit_kmeans(df, var=None, colname="cluster_id", seed=None, **kwargs):
    r"""K-means cluster a dataset

    Create a cluster-labeling model on a dataset using the K-means algorithm.

    Args:
        df (DataFrame): Hybrid point results from gr.eval_hybrid()
        var (list or None): Variables in df on which to cluster. Use None to
            cluster on all variables.
        colname (string): Name of cluster id; will be output in cluster model.
        seed (int): Random seed for kmeans clustering

    Kwargs:
        n_clusters (int): Number of clusters to fit
        random_state (int or None):

    Returns:
        gr.Model: Model that labels input data

    Notes:
        - A wrapper for sklearn.cluster.KMeans

    References:
        Scikit-learn: Machine Learning in Python, Pedregosa et al. JMLR 12, pp. 2825-2830, 2011.

    Examples:
        >>> import grama as gr
        >>> from grama.data import df_stang
        >>> from grama.fit import ft_kmeans
        >>> X = gr.Intention()
        >>> md_cluster = (
        >>>     df_stang
        >>>     >> ft_kmeans(var=["E", "mu"], n_clusters=2)
        >>> )
        >>> (
        >>>     md_cluster
        >>>     >> gr.ev_df(df_stang)
        >>>     >> gr.tf_group_by(X.cluster_id)
        >>>     >> gr.tf_summarize(
        >>>         thick_mean=gr.mean(X.thick),
        >>>         thick_sd=gr.sd(X.thick),
        >>>         n=gr.n(X.index),
        >>>     )
        >>> )

    """
    ## Check invariants
    if var is None:
        var = list(df.columns).copy()
    else:
        var = list(var).copy()
        diff = set(var).difference(set(df.columns))
        if len(diff) > 0:
            raise ValueError(
                "`var` must be subset of `df.columns`\n" "diff = {}".format(diff)
            )

    ## Generate clustering
    try:
        kmeans = KMeans(random_state=seed, **kwargs).fit(df[var].values)

    except NameError as e:
        error_string = str(e)
        raise NameError(
            error_string +
            "\n\nThis function requires the `sklearn` package. " +
            "Try running the following to install the package:\n"
            "    pip install scikit-learn"
        )

    ## Build grama model
    def fun_cluster(df):
        res = kmeans.predict(df[var].values)
        return DataFrame(data={colname: res})

    md = Model() >> cp_vec_function(fun=fun_cluster, var=var, out=[colname])

    return md


ft_kmeans = add_pipe(fit_kmeans)



__all__ = ["make_test", "df_test_input"]

from grama import cp_bounds, cp_copula_independence, cp_function, cp_marginals, \
    Model
from pandas import DataFrame


## Define a test dataset
df_test_input = DataFrame(data={"x0": [0], "x1": [0], "x2": [0]})

## Define a test model
def fun(x):
    x1, x2, x3 = x
    return x1 + x2 + x3

def make_test():
    md = Model() >> \
         cp_function(fun=fun, var=3, out=1) >> \
         cp_bounds(x0=(-1,+1), x1=(-1,+1), x2=(-1,+1)) >> \
         cp_marginals(
             x0={"dist": "uniform", "loc": -1, "scale": 2},
             x1={"dist": "uniform", "loc": -1, "scale": 2}
         ) >> \
         cp_copula_independence()

    return md



__all__ = [
    "fit_nls",
    "ft_nls",
]

## Collection of fitting synonyms: functions implemented in terms of other grama
## verbs

from grama import add_pipe, pipe, custom_formatwarning, Model, \
    df_make, tran_outer, eval_nominal, eval_nls, eval_df, \
    eval_grad_fd, cp_function, cp_md_det, cp_marginals, \
    cp_copula_gaussian, cp_bounds
from toolz import curry
from numpy import zeros, diag, atleast_2d, triu_indices
from numpy import sum as npsum
from numpy import power as nppow
from numpy import sqrt as npsqrt
from numpy.linalg import pinv, cond
from pandas import concat, DataFrame
from warnings import warn


@curry
def fit_nls(
    df_data,
    md=None,
    out=None,
    var_fix=None,
    df_init=None,
    verbose=True,
    uq_method=None,
    **kwargs,
):
    r"""Fit a model with Nonlinear Least Squares (NLS)

    Estimate best-fit variable levels with nonlinear least squares (NLS), and
    return an executable model with those frozen best-fit levels. Optionally,
    fit a distribution on the parameters to quantify parametric uncertainty.

    Note: This is a *synonym* for eval_nls(); see the documentation for
    eval_nls() for keyword argument options available beyond those listed here.

    Args:
        df_data (DataFrame): Data for estimating best-fit variable levels.
            Variables not found in df_data optimized for fitting.
        md (gr.Model): Model to analyze. All model variables
            selected for fitting must be bounded or random. Deterministic
            variables may have semi-infinite bounds.
        var_fix (list or None): Variables to fix to nominal levels. Note that
            variables with domain width zero will automatically be fixed.
        df_init (DataFrame): Initial guesses for parameters; overrides n_restart
        n_restart (int): Number of restarts to try; the first try is at
            the nominal conditions of the model. Returned model will use
            the least-error parameter set among restarts tested.
        n_maxiter (int): Optimizer maximum iterations
        verbose (bool): Print best-fit parameters to console?
        uq_method (str OR None): If string, select method to quantify parameter
            uncertainties. If None, provide best-fit values only. Methods:
            uq_method = "linpool": assume normal errors; linearly approximate
                parameter effects; equally pool variance matrices for each output

    Returns:
        gr.Model: Model for evaluation with best-fit variables frozen to
            optimized levels.

    Examples:
        >>> import grama as gr
        >>> from grama.data import df_trajectory_windowed
        >>> from grama.models import make_trajectory_linear
        >>> X = gr.Intention()
        >>>
        >>> md_trajectory = make_trajectory_linear()
        >>> md_fitted = (
        >>>     df_trajectory_windowed
        >>>     >> gr.ft_nls(
        >>>         md=md_trajectory,
        >>>         uq_method="linpool",
        >>>     )
        >>> )
    """
    ## Check `out` invariants
    if out is None:
        out = md.out
        print("... fit_nls setting out = {}".format(out))

    ## Check invariants
    if md is None:
        raise ValueError("Must provide model md")

    ## Determine variables to be fixed
    if var_fix is None:
        var_fix = set()
    else:
        var_fix = set(var_fix)
    for var in md.var_det:
        wid = md.domain.get_width(var)
        if wid == 0:
            var_fix.add(var)

    ## Run eval_nls to fit model parameter values
    df_fit = eval_nls(
        md,
        df_data=df_data,
        var_fix=var_fix,
        df_init=df_init,
        append=True,
        verbose=verbose,
        **kwargs,
    )
    ## Select best-fit values
    df_best = df_fit.sort_values(by="mse", axis=0).iloc[[0]].reset_index(drop=True)
    if verbose:
        print(df_fit.sort_values(by="mse", axis=0))

    ## Determine variables that were fitted
    var_fitted = list(set(md.var).intersection(set(df_best.columns)))
    var_remain = list(set(md.var).difference(set(var_fitted)))

    if len(var_remain) == 0:
        raise ValueError("Resulting model is constant!")

    ## Assemble and return fitted model
    if md.name is None:
        name = "(Fitted Model)"
    else:
        name = md.name + " (Fitted)"

    ## Calibrate parametric uncertainty, if requested
    if uq_method == "linpool":
        ## Precompute data
        df_nom = eval_nominal(md, df_det="nom")
        df_base = tran_outer(
            df_data, concat((df_best[var_fitted], df_nom[var_fix]), axis=1)
        )
        df_pred = eval_df(md, df=df_base)
        df_grad = eval_grad_fd(md, df_base=df_base, var=var_fitted)

        ## Pool variance matrices
        n_obs = df_data.shape[0]
        n_fitted = len(var_fitted)
        Sigma_pooled = zeros((n_fitted, n_fitted))

        for output in out:
            ## Approximate sigma_sq
            sigma_sq = npsum(
                nppow(df_data[output].values - df_pred[output].values, 2)
            ) / (n_obs - n_fitted)
            ## Approximate (pseudo)-inverse hessian
            var_grad = list(map(lambda v: "D" + output + "_D" + v, var_fitted))
            Z = df_grad[var_grad].values
            Hinv = pinv(Z.T.dot(Z), hermitian=True)

            ## Add variance matrix to pooled Sigma
            Sigma_pooled = Sigma_pooled + sigma_sq * Hinv / n_fitted

        ## Check model for identifiability
        kappa_out = cond(Sigma_pooled)
        if kappa_out > 1e10:
            warn(
                "Model is locally unidentifiable as measured by the "
                + "condition number of the pooled covariance matrix; "
                + "kappa = {}".format(kappa_out),
                RuntimeWarning,
            )

        ## Convert to std deviations and correlation
        sigma_comp = npsqrt(diag(Sigma_pooled))
        corr_mat = Sigma_pooled / (atleast_2d(sigma_comp).T.dot(atleast_2d(sigma_comp)))
        corr_data = []
        I, J = triu_indices(n_fitted, k=1)
        for ind in range(len(I)):
            i = I[ind]
            j = J[ind]
            corr_data.append([var_fitted[i], var_fitted[j], corr_mat[i, j]])
        df_corr = DataFrame(data=corr_data, columns=["var1", "var2", "corr"])

        ## Assemble marginals
        marginals = {}
        for ind, var_ in enumerate(var_fitted):
            marginals[var_] = {
                "dist": "norm",
                "loc": df_best[var_].values[0],
                "scale": sigma_comp[ind],
            }

        ## Construct model with Gaussian copula
        if len(var_fix) > 0:
            md_res = (
                Model(name)
                >> cp_function(
                    lambda x: df_nom[var_fix].values,
                    var=set(var_remain).difference(var_fix),
                    out=var_fix,
                    name="Fix variable levels",
                )
                >> cp_md_det(md=md)
                >> cp_marginals(**marginals)
                >> cp_copula_gaussian(df_corr=df_corr)
            )
        else:
            md_res = (
                Model(name)
                >> cp_md_det(md=md)
                >> cp_marginals(**marginals)
                >> cp_copula_gaussian(df_corr=df_corr)
            )

    ## Return deterministic model
    elif uq_method is None:
        md_res = (
            Model(name)
            >> cp_function(
                lambda x: df_best[var_fitted].values,
                var=var_remain,
                out=var_fitted,
                name="Fix variable levels",
            )
            >> cp_md_det(md=md)
        )

    else:
        raise ValueError("uq_method option {} not recognized".format(uq_method))

    return md_res


ft_nls = add_pipe(fit_nls)



__all__ = [
    "tran_pivot_longer",
    "tf_pivot_longer",
    "tran_pivot_wider",
    "tf_pivot_wider",
]


from grama import add_pipe, tran_select
from numpy import NaN
from pandas import DataFrame, IndexSlice, MultiIndex, RangeIndex, Series, \
    concat, isnull, pivot, pivot_table
from pandas.api.types import is_int64_dtype


def tran_pivot_longer (
    df,
    columns,
    index_to = None,
    names_to = None,
    #names_prefix = None,
    names_sep = None,
    #names_pattern = None,
    #names_ptypes = list(),
    #names_transform = list(),
    #names_repair,
    values_to = None,
    #values_drop_na = False,
    #values_ptypes = list(),
    #values_transform = list(),
):
    """

    "Lengthens" data by increasing the number of rows and decreasing the
    number of columns.

    Args:
        df (DataFrame): DataFrame passed through
        columns (str): Label of column(s) to pivot into longer format
        index_to(str): str name to create a new representation index of
                        observations
        names_to (str): name to use for the 'variable' column, if None frame.columns.name
                        is used or ‘variable’
                          • .value indicates that component of the name defines
                            the name of the column containing the cell values,
                            overriding values_to
        names_sep (str): delimter to seperate the values of the argument(s) from
                        the 'columns' parameter into 2 new columns with those
                        values split by that delimeter
                          • Regex expression is a valid input for names_sep
        values_to (str): name to use for the 'value' column

    Returns:
        DataFrame: result of being pivoted into a longer format

    Examples:

        >>> import grama as gr
        >>> from pandas import DataFrame
        >>> wide = DataFrame(
                {
                    "One": {"A": 1.0, "B": 2.0, "C": 3.0},
                    "Two": {"A": 1.0, "B": 2.0, "C": 3.0},
                }
            )
        >>> long = gr.tran_pivot_longer(
                    wide,
                    columns=("One","Two"),
                    index_to="index",
                    names_to="columns",
                    values_to="values")

    """

    ########### Pre-Check List #############
    ### Check if tran_select was used
    if isinstance(columns, DataFrame):
        columns = columns.columns.values
    ### Check if gr.tran_select helper was used:
    # NEEDS to be implmented, currently gr.tran_select needs to provided with
    # gr.matches or other helper function for this to work

    ### Check if names_to is a list or str
    names_str = False
    if isinstance(names_to, str):
        names_str = True
        if names_sep is not None:
            raise TypeError("""In order to use names_sep more than 1 value
                needs to passed to names_to""")

    ### Check for .value input
    dot_value = False
    if names_str is False:
        for i, v in enumerate(names_to):
            if names_to[i] == ".value":
                dot_value = True
    else:
        if names_to == ".value":
            dot_value = True

    ### Check values_to argument
    if values_to is None:
        values_to = "values"

    #######################################


    ########### .value pivot #############

    ### Check if .value operation needs to occur
    if dot_value is True:

        ### collect unused columns to pivot around
        data_index = collect_indexes(df, columns)

        if names_sep is not None:
            ### Add index and split column to dataset
            longer = df.reset_index().melt(
                    id_vars="index",
                    var_name="split",
                    value_vars=columns,
                    value_name=values_to
                )

            ### DataFrame Cleanup
            longer = split_cleanup(
                longer=longer,
                names_to=names_to,
                names_sep=names_sep,
                values_to=values_to
            )

        else:
            ### Add index column and .value column
            longer = df.reset_index().melt(
                    id_vars="index",
                    var_name=".value",
                    value_vars=columns,
                    value_name=values_to
                )

        ### clean up index_to call
        longer = index_to_cleanup(df, longer, data_index)

        ### arrange what indexes_from should be
        if names_str is True:
            indexes = ["index"] + data_index
        else:
            names_to = list(names_to)
            value_loc = names_to.index(".value")
            if value_loc == 0:
                indexes = ["index"] + data_index + names_to[1:]
            else:
                indexes = ["index"] + data_index + names_to[0:value_loc] \
                    + names_to[(value_loc+1):]

        ### Pivot wider the .value column
        value_longer = tran_pivot_wider(
            longer,
            indexes_from=indexes,
            names_from=".value",
            values_from=values_to
        )

        if index_to is None:
            ### drop "index" column
            value_longer.drop("index", axis=1, inplace=True)
        else:
            ### rename index column to desired: index_to
            value_longer.rename(columns={'index': index_to},inplace=True)

        return value_longer

    #########################################


    ########### names_sep pivot #############

    ### Only if names_sep is used
    if names_sep is not None:

        ### collect unused columns to pivot around
        data_index = collect_indexes(df, columns)

        if index_to is None:
            ### initial pivoted DataFrame
            longer = df.reset_index().melt(
                id_vars=data_index,
                var_name="split",
                value_vars=columns,
                value_name=values_to
            )

            ### DataFrame Cleanup
            longer = split_cleanup(
                longer=longer,
                names_to=names_to,
                names_sep=names_sep,
                values_to=values_to
            )

            return(longer)

        ### Add index column to dataset
        longer = df.reset_index().melt(
                id_vars="index",
                var_name="split",
                value_vars=columns,
                value_name=values_to
            )
        ### rename index column to desired: index_to
        longer.rename(columns={'index': index_to},inplace=True)

        longer = index_to_cleanup(df, longer, data_index)

        ### DataFrame Cleanup
        longer = split_cleanup(
            longer=longer,
            names_to=names_to,
            names_sep=names_sep,
            values_to=values_to
        )

        return(longer)

    ######################################


    ########### normal pivot #############

    ### Check if index_to is provided
    if index_to is None:

        ### check to see if all columns are used already
        data_columns = df.columns.values
        data_index = [x for x in data_columns if x not in columns]

        ### check if data_index is empty and if it has a RangeIndex
        if not data_index:
            if is_int64_dtype(df.index.dtype):
                # if so do not add extra index column and pivot
                longer = df.reset_index().melt(
                    id_vars=None,
                    var_name=names_to,
                    value_vars=columns,
                    value_name=values_to
                )
                return(longer)

            # if it does not have a RangeIndex create new column from ID column
            # and add RangeIndex
            longer = df.reset_index().melt(
                id_vars="index",
                var_name=names_to,
                value_vars=columns,
                value_name=values_to
            )
            return(longer)

        ### look for unused columns to pivot around
        data_used = columns
        data_index = [x for x in data_columns if x not in data_used]

        ### pivot with leftover name that would be the index column
        if data_index:
            longer = df.reset_index().melt(
                id_vars=data_index,
                var_name=names_to,
                value_vars=columns,
                value_name=values_to
            )
            return(longer)

    ### collect unused columns to preserve post pivot
    data_index = collect_indexes(df, columns)

    ### Add index column to dataset
    longer = df.reset_index().melt(
            id_vars="index",
            var_name=names_to,
            value_vars=columns,
            value_name=values_to
        )
    ### rename index column to desired: index_to
    longer.rename(columns={'index': index_to},inplace=True)

    longer = index_to_cleanup(df, longer, data_index)

    return longer

    ######################################

tf_pivot_longer = add_pipe(tran_pivot_longer)


def tran_pivot_wider (
     df,
     #id_cols,
     names_from,
     indexes_from = None,
     #names_prefix,
     #names_sep,
     #names_glue = None,
     #names_sort = False,
     #names_glue,
     values_from = None,
     #values_fill = None,
     #values_fn = None
 ):
    """

     "Widens" data by increasing the number of columns and decreasing the
     number of rows.

     Args:
         df (DataFrame): DataFrame passed through
         names_from (str): Column(s) name(s) to use to make new columns
         indexes_from (str): Column(s) to use to make new index, if None will
                             preserve and use unspecified columns as index
         values_from (str): Column(s) to use as new values column(s)

     Returns:
        DataFrame: result of being pivoted wider

     Example:
        >>> import grama as gr
        >>> from pandas import DataFrame
        >>> long = DataFrame(
                {
                    "index": ["A", "B", "C", "A", "B", "C"],
                    "columns": ["One", "One", "One", "Two", "Two", "Two"],
                    "values": [1.0, 2.0, 3.0, 1.0, 2.0, 3.0],
                }
            )
        >>> wide = gr.tran_pivot_wider(
                long,
                indexes_from="index",
                names_from="columns",
                values_from="values"
            )

    """

    if indexes_from is None:
        ### Clean columns list to find unused columns and preserve them
        data_columns = df.columns
        data_values_used = values_from
        data_columns_used = names_from
        data_1st_clean = [x for x in data_columns if x not in data_columns_used]
        data_index_clean = [x for x in data_1st_clean if x not in data_values_used]
        if not data_index_clean:
            data_index_clean = None

        ### try pivot, if no repeats exist in data_index_clean values continue,
        ### else use pivot_table to handle repreated values
        wider = pivot(
            df,
            index=data_index_clean,
            columns=names_from,
            values=values_from
        )
        ### Post cleaning of leftover column and index names
        wider.columns.name = None
        # preserve rows by re-inserting them
        if data_index_clean is not None:
            for i, v in enumerate(data_index_clean):
                wider.insert(
                    loc=i,
                    column=data_index_clean[i],
                    value=wider.index.get_level_values(data_index_clean[i])
                    )
        # remake index to be numbered column with no name
        wider.index = RangeIndex(start=0,stop=len(wider),step=1)
        wider.index.name = None

        return wider

    ### If indexes_from exists do the same repeat check as above but with the
    ### provided indexes
    wider = pivot(
        df,
        index=indexes_from,
        columns=names_from,
        values=values_from
    )
    ### Post cleaning of leftover column and index names
    wider.columns.name = None
    ## preserve rows by re-inserting them
    if isinstance(indexes_from, str):
        wider.insert(0,indexes_from,wider.index)
    # if str, insert as so, else loop through list type of indexes_from
    else:
        for i, v in enumerate(indexes_from):
            wider.insert(
                loc=i,
                column=indexes_from[i],
                value=wider.index.get_level_values(indexes_from[i])
                )
    # remake index to be numbered column with no name
    wider.index = RangeIndex(start=0,stop=len(wider),step=1)
    wider.index.name = None

    return wider

tf_pivot_wider = add_pipe(tran_pivot_wider)


def split_cleanup(
    longer,
    names_to,
    names_sep,
    values_to
):
    """
        split_cleanup cleans up pivots that use the names_sep functionality
    """
    ### clean up DataFrame
    # split columns
    split_columns = longer.split.str.split(
        names_sep,
        expand=True,
        n=0
    )
    # add back split columns to DataFrame
    longer = concat([longer,split_columns], axis=1)
    # drop column that the split came from
    longer.drop("split", axis=1, inplace=True)
    # rename columns
    for i,v in enumerate(names_to):
        longer.rename(columns={i: names_to[i]},inplace=True)
    # if any values are None make them NaN
    for i,v in enumerate(names_to):
        for j, w in enumerate(longer[names_to[i]]):
            if w is None:
                longer[names_to[i]][j] = NaN
    # reorder values column to the end
    longer = longer[[c for c in longer if c not in values_to] + [values_to]]

    return(longer)


def collect_indexes(df, columns):
    """
        collect_indexes finds unused column to pivot around
    """
    ### look for unused columns to pivot around
    data_used = columns
    data_columns = df.columns.values
    data_index = [x for x in data_columns if x not in data_used]

    return(data_index)

def index_to_cleanup(df, longer, data_index):
    """
        index_to_cleanup cleansup longer if index_to was called to the function
    """
    ### if there was columns needing to be re-inserted do so
    if data_index is not None:
        for i, v in enumerate(data_index):
            longer.insert(
                loc=i+1,
                column=data_index[i],
                value=df[data_index[i]]
            )

    ### repair NaN values from transformation
    for index in data_index:
        length = len(df[index])
        for i, v in enumerate(longer[index]):
            if isnull(longer[index][i]):
                longer[index][i] = longer[index][i%length]

    return(longer)



__all__ = [
    "fit_lolo",
    "ft_lolo",
]

## Fitting via sklearn package
try:
    from lolopy.learners import RandomForestRegressor
except ModuleNotFoundError:
    pass

from grama import add_pipe, Function, Model, pipe
from numpy import stack
from numpy.random import seed as set_seed
from pandas import DataFrame
from toolz import curry
from warnings import filterwarnings


## Helper functions and classes
# --------------------------------------------------
class FunctionRFR(Function):
    def __init__(self, rf, var, out, name, runtime, return_std):
        """

        Args:
            rf (scikit RandomForestRegressor):
        """
        self.rf = rf
        self.var = var
        self.name = name
        self.runtime = runtime

        self.return_std = return_std
        if return_std:
            self.out = list(map(lambda s: s + "_mean", out)) + list(map(lambda s: s + "_sd", out))
        else:
            self.out = list(map(lambda s: s + "_mean", out))

    def eval(self, df):
        ## Check invariant; model inputs must be subset of df columns
        if not set(self.var).issubset(set(df.columns)):
            raise ValueError(
                "Model function `{}` var not a subset of given columns".format(
                    self.name
                )
            )

        ## Predict
        if self.return_std:
            res = self.rf.predict(df[self.var], return_std=True)
            y = stack((res[0], res[1])).T
        else:
            y = self.rf.predict(df[self.var])

        return DataFrame(data=y, columns=self.out)


## Fit random forest model with sklearn
# --------------------------------------------------
@curry
def fit_lolo(
    df,
    md=None,
    var=None,
    out=None,
    domain=None,
    density=None,
    seed=None,
    return_std=True,
    suppress_warnings=True,
    **kwargs
):
    r"""Fit a random forest

    Fit a random forest to given data. Specify inputs and outputs, or inherit
    from an existing model.

    Args:
        df (DataFrame): Data for function fitting
        md (gr.Model): Model from which to inherit metadata
        var (list(str) or None): List of features or None for all except outputs
        out (list(str)): List of outputs to fit
        domain (gr.Domain): Domain for new model
        density (gr.Density): Density for new model
        seed (int or None): Random seed for fitting process
        return_std (bool): Return predictive standard deviations?
        suppress_warnings (bool): Suppress warnings when fitting?

    Keyword Arguments:

        num_trees (int):
        use_jackknife (bool):
        bias_learner ():
        leaf_learner ():
        subset_strategy (str):
        min_leaf_instances (int):
        max_depth (int):
        uncertainty_calibration (bool):
        randomize_pivot_location (bool):
        randomly_rotate_features (bool):

    Returns:
        gr.Model: A grama model with fitted function(s)

    Notes:
        - Wrapper for lolopy.learners.RandomForestRegressor

    """
    if suppress_warnings:
        filterwarnings("ignore")

    n_obs, n_in = df.shape

    ## Check minimum rows
    if n_obs < 8:
        raise ValueError("The lolo random forest requires at least 8 rows")

    ## Infer fitting metadata, if available
    if not (md is None):
        domain = md.domain
        density = md.density
        out = md.out

    ## Check invariants
    if not set(out).issubset(set(df.columns)):
        raise ValueError("out must be subset of df.columns")
    ## Default input value
    if var is None:
        var = list(set(df.columns).difference(set(out)))
    ## Check more invariants
    set_inter = set(out).intersection(set(var))
    if len(set_inter) > 0:
        raise ValueError(
            "outputs and inputs must be disjoint; intersect = {}".format(set_inter)
        )
    if not set(var).issubset(set(df.columns)):
        raise ValueError("var must be subset of df.columns")

    ## Construct gaussian process for each output
    functions = []

    try:
        for output in out:
            rf = RandomForestRegressor(**kwargs)
            set_seed(seed)
            rf.fit(df[var].values, df[output].values)
            name = "RF"

            fun = FunctionRFR(rf, var, [output], name, 0, return_std)
            functions.append(fun)

    except NameError as e:
        error_string = str(e)
        raise NameError(
            error_string +
            "\n\nThis function requires the `lolopy` package. " +
            "Try running the following to install the package:\n"
            "    pip install lolopy"
        )

    ## Construct model
    return Model(functions=functions, domain=domain, density=density)


ft_lolo = add_pipe(fit_lolo)



__all__ = [
    "plot_scattermat",
    "pt_scattermat",
    "plot_hists",
    "pt_hists",
    "plot_sinew_inputs",
    "pt_sinew_inputs",
    "plot_sinew_outputs",
    "pt_sinew_outputs",
    "plot_auto",
    "pt_auto",
    "plot_list",
]

from grama import add_pipe, pipe, tf_pivot_longer, tf_outer, tf_select, tf_rename, tf_filter
from grama import Intention
from pandas import melt

from plotnine import aes, annotate, ggplot, facet_grid, facet_wrap, labs, element_text, guides
from plotnine import theme, theme_void, theme_minimal
from plotnine import scale_x_continuous, scale_y_continuous
from plotnine import geom_point, geom_density, geom_histogram, geom_line, geom_blank
from matplotlib import gridspec

from toolz import curry

## Helper functions
##################################################
def _sci_format(v):
    r"""Scientific format
    """
    return (
        ["{0:1.1e}".format(v[0])] +
        [""] * (len(v) - 2) +
        ["{0:1.1e}".format(v[-1])]
    )

## Function-specific plot functions
##################################################
## Monte Carlo
# --------------------------------------------------
@curry
def plot_scattermat(df, var=None):
    r"""Create a scatterplot matrix

    Create a scatterplot matrix. Often used to visualize a design (set of inputs
    points) before evaluating the functions.

    Usually called as a dispatch from plot_auto().

    Args:
        var (list of strings): Variables to plot

    Returns:
        DataFrame: Results of evaluation or unevaluated design

    Examples:

        >>> import grama as gr
        >>> import matplotlib.pyplot as plt
        >>> from grama.models import make_cantilever_beam
        >>> md = make_cantilever_beam()
        >>> ## Dispatch from autoplotter
        >>> (
        >>>     md
        >>>     >> gr.ev_sample(n=100, df_det="nom", skip=True)
        >>>     >> gr.pt_auto()
        >>> )
        >>> ## Re-create plot without metadata
        >>> (
        >>>     md
        >>>     >> gr.ev_sample(n=100, df_det="nom")
        >>>     >> gr.pt_scattermat(var=md.var)
        >>> )

    """
    if var is None:
        raise ValueError("Must provide input columns list as keyword var")

    ## Define helpers
    labels_blank = lambda v: [""] * len(v)
    breaks_min = lambda lims: (lims[0], 0.5 * (lims[0] + lims[1]), lims[1])

    ## Make blank figure
    fig = (
        df
        >> ggplot()
        + geom_blank()
        + theme_void()
    ).draw(show=False)

    gs = gridspec.GridSpec(len(var), len(var))
    for i, v1 in enumerate(var):
        for j, v2 in enumerate(var):
            ax = fig.add_subplot(gs[i, j])
            ## Switch labels
            if j == 0:
                labels_y = _sci_format
            else:
                labels_y = labels_blank
            if i == len(var) - 1:
                labels_x = _sci_format
            else:
                labels_x = labels_blank

            ## Density
            if i == j:
                xmid = 0.5 * (
                    df[v1].min() + df[v1].max()
                )

                p = (
                    df
                    >> ggplot(aes(v1))
                    + geom_density()
                    + scale_x_continuous(
                        breaks=breaks_min,
                        labels=labels_x,
                    )
                    + scale_y_continuous(
                        breaks=breaks_min,
                        labels=labels_y,
                    )
                    + annotate(
                        "label",
                        x=xmid,
                        y=0,
                        label=v1,
                        va="bottom",
                    )
                    + theme_minimal()
                    + labs(title=v1)
                )

            ## Scatterplot
            else:
                p = (
                    df
                    >> ggplot(aes(v2, v1))
                    + geom_point()
                    + scale_x_continuous(
                        breaks=breaks_min,
                        labels=labels_x,
                    )
                    + scale_y_continuous(
                        breaks=breaks_min,
                        labels=labels_y,
                    )
                    + theme_minimal()
                    + theme(
                        axis_title=element_text(va="top", size=12),
                    )
                )

            _ = p._draw_using_figure(fig, [ax])


    ## Plot
    # NB Returning the figure causes a "double plot" in Jupyter....
    fig.show()


pt_scattermat = add_pipe(plot_scattermat)


@curry
def plot_hists(df, out=None):
    r"""Construct histograms

    Create a set of histograms. Often used to visualize the results of random
    sampling for multiple outputs.

    Usually called as a dispatch from plot_auto().

    Args:
        out (list of strings): Variables to plot

    Returns:
        Seaborn histogram plot

    Examples:

        >>> import grama as gr
        >>> import matplotlib.pyplot as plt
        >>> from grama.models import make_cantilever_beam
        >>> md = make_cantilever_beam()
        >>> ## Dispatch from autoplotter
        >>> (
        >>>     md
        >>>     >> gr.ev_sample(n=100, df_det="nom")
        >>>     >> gr.pt_auto()
        >>> )
        >>> ## Re-create without metadata
        >>> (
        >>>     md
        >>>     >> gr.ev_sample(n=100, df_det="nom")
        >>>     >> gr.pt_hists(out=md.out)
        >>> )

    """
    if out is None:
        raise ValueError("Must provide input columns list as keyword out")

    return (
        df
        >> tf_pivot_longer(
            columns=out,
            names_to="var",
            values_to="value",
        )
        >> ggplot(aes("value"))
        + geom_histogram(bins=30)
        + facet_wrap("var", scales="free")
        + theme_minimal()
        + labs(
            x="Output Value",
            y="Count",
        )
    )


pt_hists = add_pipe(plot_hists)

## Sinew plots
# --------------------------------------------------
@curry
def plot_sinew_inputs(df, var=None, sweep_ind="sweep_ind"):
    r"""Inspect a sinew design

    Create a scatterplot matrix with hues. Often used to visualize a sinew
    design before evaluating the model functions.

    Usually called as a dispatch from plot_auto().

    Args:
        df (Pandas DataFrame): Input design data
        var (list of strings): Variables to plot
        sweep_ind (string): Sweep index column in df

    Returns:
        Seaborn scatterplot matrix

    Examples:

        >>> import grama as gr
        >>> import matplotlib.pyplot as plt
        >>> from grama.models import make_cantilever_beam
        >>> md = make_cantilever_beam()
        >>> ## Dispatch from autoplotter
        >>> (
        >>>     md
        >>>     >> gr.ev_sinews(df_det="swp", skip=True)
        >>>     >> gr.pt_auto()
        >>> )
        >>> ## Re-create without metadata
        >>> (
        >>>     md
        >>>     >> gr.ev_sinews(df_det="swp")
        >>>     >> gr.pt_sinew_inputs(var=md.var)
        >>> )

    """
    if var is None:
        raise ValueError("Must provide input columns list as keyword var")

    ## Define helpers
    labels_blank = lambda v: [""] * len(v)
    breaks_min = lambda lims: (lims[0], 0.5 * (lims[0] + lims[1]), lims[1])

    ## Make blank figure
    fig = (
        df
        >> ggplot()
        + geom_blank()
        + theme_void()
    ).draw(show=False)

    gs = gridspec.GridSpec(len(var), len(var))
    for i, v1 in enumerate(var):
        for j, v2 in enumerate(var):
            ax = fig.add_subplot(gs[i, j])
            ## Switch labels
            if j == 0:
                labels_y = _sci_format
            else:
                labels_y = labels_blank
            if i == len(var) - 1:
                labels_x = _sci_format
            else:
                labels_x = labels_blank

            ## Label
            if i == j:
                p = (
                    df
                    >> ggplot()
                    + annotate(
                        "label",
                        x=0,
                        y=0,
                        label=v1,
                    )
                    + theme_void()
                    + guides(color=None)
                )

            ## Scatterplot
            else:
                p = (
                    df
                    >> ggplot(aes(v2, v1, color="factor("+sweep_ind+")"))
                    + geom_point(size=0.1)
                    + scale_x_continuous(
                        breaks=breaks_min,
                        labels=labels_x,
                    )
                    + scale_y_continuous(
                        breaks=breaks_min,
                        labels=labels_y,
                    )
                    + guides(color=None)
                    + theme_minimal()
                    + theme(
                        axis_title=element_text(va="top", size=12),
                    )
                )

            _ = p._draw_using_figure(fig, [ax])


    ## Plot
    # NB Returning the figure causes a "double plot" in Jupyter....
    fig.show()


pt_sinew_inputs = add_pipe(plot_sinew_inputs)


@curry
def plot_sinew_outputs(
    df, var=None, out=None, sweep_ind="sweep_ind", sweep_var="sweep_var"
):
    r"""Construct sinew plot

    Create a relational lineplot with hues. Often used to visualize the outputs
    of a sinew design.

    Usually called as a dispatch from plot_auto().

    Args:
        df (Pandas DataFrame): Input design data with output results
        var (list of strings): Variables to plot
        out (list of strings): Outputs to plot
        sweep_ind (string): Sweep index column in df
        sweep_var (string): Swept variable column in df

    Returns:
        Seaborn relational lineplot

    Examples:

        >>> import grama as gr
        >>> import matplotlib.pyplot as plt
        >>> from grama.models import make_cantilever_beam
        >>> md = make_cantilever_beam()
        >>> ## Dispatch from autoplotter
        >>> (
        >>>     md
        >>>     >> gr.ev_sinews(df_det="swp")
        >>>     >> gr.pt_auto()
        >>> )
        >>> ## Re-create without metadata
        >>> (
        >>>     md
        >>>     >> gr.ev_sinews(df_det="swp")
        >>>     >> gr.pt_sinew_inputs(var=md.var, out=md.out)
        >>> )

    """
    if var is None:
        raise ValueError("Must provide input columns list as keyword arg var")
    if out is None:
        raise ValueError("Must provide output columns list as keyword arg out")

    ## Prepare data
    # Gather inputs
    id_vars = [col for col in df.columns if col not in var]
    df_tmp = melt(df, id_vars, var, "_var", "_x")

    # Gather outputs
    id_vars = [col for col in df_tmp.columns if col not in out]
    df_plot = melt(df_tmp, id_vars, out, "_out", "_y")

    # Filter off-sweep values
    df_plot = df_plot[df_plot[sweep_var] == df_plot["_var"]]

    breaks_min = lambda lims: (lims[0], 0.5 * (lims[0] + lims[1]), lims[1])
    return (
        df_plot
        >> ggplot(aes(
            "_x",
            "_y",
            color="factor(" + sweep_ind + ")",
            group="factor(" + sweep_ind + ")",
        ))
        + geom_line()
        + facet_grid("_out~_var", scales="free")

        + scale_x_continuous(
            breaks=breaks_min,
            labels=_sci_format,
        )
        + scale_y_continuous(
            breaks=breaks_min,
            labels=_sci_format,
        )
        + guides(color=None)
        + theme_minimal()
        + labs(
            x="Input Value",
            y="Output Value",
        )
    )


pt_sinew_outputs = add_pipe(plot_sinew_outputs)

## Autoplot dispatcher
## ##################################################
plot_list = {
    "sinew_inputs": plot_sinew_inputs,
    "sinew_outputs": plot_sinew_outputs,
    "sample_inputs": plot_scattermat,
    "sample_outputs": plot_hists,
}


@curry
def plot_auto(df):
    r"""Automagic plotting

    Convenience tool for various grama outputs. Prints delegated plotting
    function, which can be called manually with different arguments for
    more tailored plotting.

    Args:
        df (DataFrame): Data output from appropriate grama routine. See
            gr.plot_list.keys() for list of supported methods.

    Returns:
        Plot results

    """
    try:
        d = df._plot_info
    except AttributeError:
        raise AttributeError(
            "'{}' object has no attribute _plot_info. Use plot_auto() for grama outputs only.".format(
                "DataFrame"
            )
        )

    try:
        plot_fcn = plot_list[d["type"]]
    except KeyError:
        raise KeyError("'{}' Plot type not implemented.".format(d["type"]))
    plt_kwargs = {key: value for key, value in d.items() if key != "type"}

    print("Calling {0:}....".format(plot_fcn.__name__))

    return plot_fcn(df, **plt_kwargs)


pt_auto = add_pipe(plot_auto)



__all__ = [
    "str_c",
    "str_count",
    "str_detect",
    "str_extract",
    "str_locate",
    "str_replace",
    "str_replace_all",
    "str_sub",
    "str_split",
    "str_which",
    "str_to_lower",
    "str_to_upper",
    "str_to_snake",
]

import re
from grama import make_symbolic, pipe, valid_dist, param_dist, dfdelegate
from numpy import NaN, isnan
from pandas import Series


# ------------------------------------------------------------------------------
# String helpers
# - a straight port of stringr
# ------------------------------------------------------------------------------

## Vectorized Concatenate
# --------------------------------------------------
def _vec_len(x):
    try:
        if isinstance(x, str):
            raise TypeError
        return len(x)
    except TypeError:
        return 1


def _ensure_series(x, l, length):
    if (l == 1) and (not isinstance(x, Series)):
        return Series([x] * length).astype(str)

    return Series(x).astype(str).reset_index(drop=True)


@make_symbolic
def str_c(*args, sep=""):
    """
    Concatenate strings
    """
    ## Pass-through a single arg
    if len(args) < 2:
        return Series(args[0]).astype(str)

    ## Check all lengths 1 or equal
    all_lengths = [_vec_len(a) for a in args]
    if not len(set(filter(lambda l: l > 1, all_lengths))) <= 1:
        raise ValueError("All arguments must be same length or scalar")
    length = max(all_lengths)

    ## Ensure first arg is string
    res = _ensure_series(args[0], all_lengths[0], length)

    ## Iteratively concatenate
    for i in range(1, len(args)):
        tmp = _ensure_series(args[i], all_lengths[i], length)
        res = res.str.cat(tmp, sep=sep)

    return res


## Detect matches
# --------------------------------------------------
@make_symbolic
def str_detect(string, pattern):
    """
    Detect the presence of a pattern match in a string.
    """
    try:
        ## Escape by raise if string is single string
        if isinstance(string, str):
            raise TypeError
        return Series([not (re.search(pattern, s) is None) for s in string])

    ## Single string
    except TypeError:
        return not (re.search(pattern, string) is None)


@make_symbolic
def str_locate(string, pattern):
    """
    Find the indices of all pattern matches.
    """
    try:
        if isinstance(string, str):
            raise TypeError
        return Series([[m.start(0) for m in re.finditer(pattern, s)] for s in string])

    except TypeError:
        return [m.start(0) for m in re.finditer(pattern, string)]


def _safe_index(l, i=0):
    try:
        return l[i]
    except IndexError:
        return NaN


@make_symbolic
def str_which(string, pattern):
    """
    Find the index of the first pattern match.
    """
    indices = str_locate(string, pattern)

    try:
        if isinstance(string, str):
            raise TypeError
        return Series([_safe_index(inds) for inds in indices])

    except TypeError:
        return _safe_index(indices)


@make_symbolic
def str_count(string, pattern):
    """
    Count the number of matches in a string.
    """
    indices = str_locate(string, pattern)

    try:
        if isinstance(string, str):
            raise TypeError
        return Series([len(ind) for ind in indices])

    except TypeError:
        return len(indices)


## Subset strings
# --------------------------------------------------
@make_symbolic
def str_sub(string, start=0, end=None):
    """Extract substrings"""
    try:
        if isinstance(string, str):
            raise TypeError
        return Series([s[start:end] for s in string])

    except TypeError:
        return string[start:end]


def _extract_or_empty(string, pattern):
    match = re.search(pattern, string)

    try:
        return match.group(0)
    except AttributeError:
        return ""


@make_symbolic
def str_extract(string, pattern):
    """Extract the first regex pattern match"""

    try:
        if isinstance(string, str):
            raise TypeError
        return Series([_extract_or_empty(s, pattern) for s in string])

    except TypeError:

        return _extract_or_empty(string, pattern)


## Mutate string
# --------------------------------------------------
@make_symbolic
def str_replace(string, pattern, replacement):
    """Replace the first matched pattern in each string."""

    try:
        if isinstance(string, str):
            raise TypeError
        return Series([re.sub(pattern, replacement, s, count=1) for s in string])

    except TypeError:

        return re.sub(pattern, replacement, string, count=1)


@make_symbolic
def str_replace_all(string, pattern, replacement):
    """Replace all occurences of pattern in each string."""

    try:
        if isinstance(string, str):
            raise TypeError
        return Series([re.sub(pattern, replacement, s) for s in string])

    except TypeError:

        return re.sub(pattern, replacement, string)


## Split
# --------------------------------------------------
@make_symbolic
def str_split(string, pattern, maxsplit=0):
    """Split string into list on pattern

    Args:
        string (str or iterable[str]): String(s) to split
        pattern (str): Regex pattern on which to split

    Kwargs:
        maxsplit (int): Maximum number of splits, or 0 for unlimited

    Returns
        str or iterable[str]: List (of lists) of strings

    """

    try:
        if isinstance(string, str):
            raise TypeError
        return Series([re.split(pattern, s, maxsplit=maxsplit) for s in string])

    except TypeError:

        return re.split(pattern, string, maxsplit=maxsplit)


## Case helpers
# --------------------------------------------------
@make_symbolic
def str_to_lower(string):
    """Make string lower-case

    Args:
        string (str or iterable[str]): String(s) to modify

    Returns
        str or iterable[str]: List (of lists) of strings

    """

    try:
        if isinstance(string, str):
            raise TypeError
        return Series([s.lower() for s in string])

    except TypeError:

        return string.lower()


@make_symbolic
def str_to_upper(string):
    """Make string upper-case

    Args:
        string (str or iterable[str]): String(s) to modify

    Returns
        str or iterable[str]: List (of lists) of strings

    """

    try:
        if isinstance(string, str):
            raise TypeError
        return Series([s.upper() for s in string])

    except TypeError:

        return string.upper()


def _snake_case(s):
    return re.sub("\s+", "_", s)


@make_symbolic
def str_to_snake(string):
    """Make string snake case

    Args:
        string (str or iterable[str]): String(s) to modify

    Returns
        str or iterable[str]: List (of lists) of strings

    """

    try:
        if isinstance(string, str):
            raise TypeError
        return Series([_snake_case(s) for s in string])

    except TypeError:

        return _snake_case(string)



__all__ = [
    "tran_tsne",
    "tf_tsne",
    "tran_poly",
    "tf_poly",
]

## Transforms via sklearn package
try:
    from sklearn.manifold import TSNE
    from sklearn.preprocessing import PolynomialFeatures

except ModuleNotFoundError:
    pass

from grama import add_pipe
from pandas import concat, DataFrame
from toolz import curry


## Compute t-SNE
# --------------------------------------------------
@curry
def tran_tsne(
    df, var=None, out="xi", keep=True, append=False, n_dim=2, seed=None, **kwargs
):
    r"""t-SNE dimension reduction of a dataset

    Apply the t-SNE algorithm to reduce the dimensionality of a dataset.

    Args:
        df (DataFrame): Hybrid point results from gr.eval_hybrid()
        var (list or None): Variables in df on which to perform dimension reduction.
            Use None to compute with all variables.
        out (string): Name of reduced-dimensionality output; indexed from 0 .. n_dim-1
        keep (bool): Keep unused columns (outside `var`) in new DataFrame?
        append (bool): Append results to original columns?
        n_dim (int): Target dimensionality

    Kwargs:
        n_iter (int): Maximum number of iterations for optimization. As Wattenberg et al. note, this is the most important parameter in using t-SNE. If you see strange "pinched" shapes, increase n_iter.
        perplexity (int): Usually between 5 and 50. Low perplexity means local variations dominate; High perplexity tends to merge clusters.
        early_exaggeration (float):
        learning_rate (float):

    Notes:
        - A wrapper for sklearn.manifold.TSNE

    References:
        Scikit-learn: Machine Learning in Python, Pedregosa et al. JMLR 12, pp. 2825-2830, 2011.

        Wattenberg, Viegas, and Johnson, "How to use t-SNE effectively" (2016) Distil.pub

    Examples:

    """
    ## Check invariants
    if var is None:
        var = list(df.columns).copy()
    else:
        var = list(var).copy()
        diff = set(var).difference(set(df.columns))
        if len(diff) > 0:
            raise ValueError(
                "`var` must be subset of `df.columns`\n" "diff = {}".format(diff)
            )
    var_leftover = list(set(df.columns).difference(set(var)))

    ## Reduce dimensionality
    try:
        df_res = DataFrame(
            data=TSNE(
                n_components=n_dim,
                random_state=seed,
                **kwargs
            ).fit_transform(df[var].values),
            columns=[out + "{}".format(i) for i in range(n_dim)],
        )

    except NameError as e:
        error_string = str(e)
        raise NameError(
            error_string +
            "\n\nThis function requires the `sklearn` package. " +
            "Try running the following to install the package:\n"
            "    pip install scikit-learn"
        )

    ## Concatenate as necessary
    if keep:
        df_res = concat(
            (df_res.reset_index(drop=True), df[var_leftover].reset_index(drop=True)),
            axis=1,
        )
    if append:
        df_res = concat(
            (df_res.reset_index(drop=True), df[var].reset_index(drop=True)), axis=1
        )

    return df_res


tf_tsne = add_pipe(tran_tsne)

# --------------------------------------------------
@curry
def tran_poly(df, degree=None, var=None, keep=True, **kwargs):
    r"""Compute polynomial features of a dataset

    Compute polynomial features of a dataset.

    Args:
        df (DataFrame): Hybrid point results from gr.eval_hybrid()

    Kwargs:
        degree (int): Maximum degree of polynomial features
        var (list or None): Variables in df on which to perform dimension reduction.
            Use None to compute with all variables.
        keep (bool): Keep unused columns (outside `var`) in new DataFrame?
        interaction_only (bool): If true, only produce interaction features
        include_bias (bool): If true, include a constant feature term (bias)

    Notes:
        - A wrapper for sklearn.preprocessing.PolynomialFeatures

    References:
        Scikit-learn: Machine Learning in Python, Pedregosa et al. JMLR 12, pp. 2825-2830, 2011.

    Examples:

    """
    ## Check invariants
    if var is None:
        var = list(df.columns).copy()
    else:
        var = list(var).copy()
        diff = set(var).difference(set(df.columns))
        if len(diff) > 0:
            raise ValueError(
                "`var` must be subset of `df.columns`\n" "diff = {}".format(diff)
            )
    var_leftover = list(set(df.columns).difference(set(var)))

    ## Compute the features
    try:
        fit = PolynomialFeatures(degree)

    except NameError as e:
        error_string = str(e)
        raise NameError(
            error_string +
            "\n\nThis function requires the `sklearn` package. " +
            "Try running the following to install the package:\n"
            "    pip install scikit-learn"
        )

    X_feat = fit.fit_transform(df[var].values)
    var_feat = fit.get_feature_names(var)

    ## Package results
    df_feat = DataFrame(data=X_feat, columns=var_feat)

    if keep:
        return concat((df_feat, df[var_leftover].reset_index(drop=True)), axis=1)

    return df_feat


tf_poly = add_pipe(tran_poly)



__all__ = ["make_prlc", "make_prlc_rand"]

from grama import cp_bounds, cp_copula_independence, cp_function, \
    cp_bounds, cp_marginals, cp_vec_function, df_make, Model
from numpy import sqrt


## Component tolerances
C_percent_lo = -0.20
C_percent_up = +0.80
R_percent_lo = -0.05
R_percent_up = +0.05
L_percent_lo = -0.10
L_percent_up = +0.10

def make_prlc():
    md_RLC_det = (
        Model("RLC Circuit")
        >> cp_vec_function(
            fun=lambda df: df_make(omega0=sqrt(1 / df.L / df.C)),
            var=["L", "C"],
            out=["omega0"],
        )
        >> cp_function(
            fun=lambda df: df_make(Q=df.omega0 * df.R * df.C),
            name="parallel RLC",
            var=["omega0", "R", "C"],
            out=["Q"]
        )
        >> cp_bounds(
            R=(1e-3, 1e0),
            L=(1e-9, 1e-3),
            C=(1e-3, 100),
        )
    )

    return md_RLC_det

def make_prlc_rand():
    md_RLC_rand = (
        Model("RLC with component tolerances")
        >> cp_vec_function(
            fun=lambda df: df_make(
                Rr=df.R * (1 + df.dR),
                Lr=df.L * (1 + df.dL),
                Cr=df.C * (1 + df.dC),
            ),
            var=["R", "dR", "L", "dL", "C", "dC"],
            out=["Rr", "Lr", "Cr"],
        )
        >> cp_vec_function(
            fun=lambda df: df_make(
                omega0=sqrt(1 / df.Lr / df.Cr)
            ),
            var=["Lr", "Cr"],
            out=["omega0"],
        )
        >> cp_vec_function(
            fun=lambda df: df_make(Q=df.omega0 * df.Rr * df.Cr),
            name="parallel RLC",
            var=["omega0", "Rr", "Cr"],
            out=["Q"]
        )
        >> cp_bounds(
            R=(1e-3, 1e0),
            L=(1e-9, 1e-3),
            C=(1e-3, 100),
        )
        >> cp_marginals(
            dR=dict(
                dist="uniform",
                loc=R_percent_lo,
                scale=R_percent_up - R_percent_lo
            ),
            dL=dict(
                dist="uniform",
                loc=L_percent_lo,
                scale=L_percent_up - L_percent_lo
            ),
            dC=dict(
                dist="uniform",
                loc=C_percent_lo,
                scale=C_percent_up - C_percent_lo
            ),
        )
        >> cp_copula_independence()
    )

    return md_RLC_rand



__all__ = [
    "str_detect",
    "str_count",
    "str_locate",
    "str_replace",
    "str_which",
]

import re
from .. import dfcurry, pipe, valid_dist, param_dist
from numpy import NaN, isnan
from pandas import Series


# ------------------------------------------------------------------------------
# String helpers
# - a straight port of stringr
# ------------------------------------------------------------------------------

## Detect matches
# --------------------------------------------------
@make_symbolic
def str_detect(string, pattern):
    """
    Detect the presence of a pattern match in a string.
    """
    try:
        ## Escape by raise if string is single string
        if isinstance(string, str):
            raise TypeError
        return Series([not (re.search(pattern, s) is None) for s in string])

    ## Single string
    except TypeError:
        return not (re.search(pattern, string) is None)


@make_symbolic
def str_locate(string, pattern):
    """
    Find the indices of all pattern matches.
    """
    try:
        if isinstance(string, str):
            raise TypeError
        return Series([[m.start(0) for m in re.finditer(pattern, s)] for s in string])

    except TypeError:
        return [m.start(0) for m in re.finditer(pattern, string)]


def _safe_index(l, ind=0):
    try:
        return l[ind]
    except IndexError:
        return NaN


@make_symbolic
def str_which(string, pattern):
    """
    Find the index of the first pattern match.
    """
    indices = str_locate(string, pattern)

    try:
        if isinstance(string, str):
            raise TypeError
        return Series([_safe_index(ind) for ind in indices])

    except TypeError:
        return _safe_index(indices)


@make_symbolic
def str_count(string, pattern):
    """
    Count the number of matches in a string.
    """
    indices = str_locate(string, pattern)

    try:
        if isinstance(string, str):
            raise TypeError
        return Series([len(ind) for ind in indices])

    except TypeError:
        return len(indices)


## Subset strings
# --------------------------------------------------

## Mutate string
# --------------------------------------------------
def replace_or_none(string, pattern, replacement, ind):
    if not isnan(ind):
        return string[:ind] + replacement + string[ind + len(pattern) :]
    return string


@make_symbolic
def str_replace(string, pattern, replacement):
    """Replace the first matched pattern in each string."""

    indices = str_which(string, pattern)

    try:
        if isinstance(string, str):
            raise TypeError
        return Series(
            [
                replace_or_none(string[ind], pattern, replacement, indices[ind])
                for ind in range(len(indices))
            ]
        )

    except TypeError:

        return replace_or_none(string, pattern, replacement, indices)



__all__ = [
    "marg_gkde",
    "marg_named",
    "marg_mom",
    "Marginal",
    "MarginalNamed",
    "MarginalGKDE",
    "param_dist",
    "valid_dist",
]

import copy
import warnings
from grama import make_symbolic
from abc import ABC, abstractmethod
from numpy import zeros, array, Inf, concatenate, sqrt
from numpy import min as npmin
from numpy import max as npmax
from pandas import DataFrame
from scipy.optimize import root_scalar, root
from scipy.stats import alpha, anglit, arcsine, argus, beta, betaprime, \
    bradford, burr, burr12, cauchy, chi, chi2, cosine, crystalball, dgamma, \
    dweibull, erlang, expon, exponnorm, exponweib, exponpow, f, fatiguelife, \
    fisk, foldcauchy, foldnorm, gaussian_kde, genlogistic, gennorm, genpareto, \
    genexpon, genextreme, gausshyper, gamma, gengamma, genhalflogistic, \
    gilbrat, gompertz, gumbel_r, gumbel_l, halfcauchy, halflogistic, \
    halfnorm, halfgennorm, hypsecant, invgamma, invgauss, invweibull, \
    johnsonsb, johnsonsu, kappa4, kappa3, ksone, kstwobign, laplace, levy, \
    levy_l, levy_stable, logistic, loggamma, loglaplace, lognorm, lomax, \
    maxwell, mielke, moyal, nakagami, ncx2, ncf, nct, norm, norminvgauss, \
    pareto, pearson3, powerlaw, powerlognorm, powernorm, rdist, rayleigh, \
    rice, recipinvgauss, skewnorm, t, trapz, triang, truncexpon, truncnorm, \
    tukeylambda, uniform, vonmises, vonmises_line, wald, weibull_min, \
    weibull_max, wrapcauchy


## Scipy metadata
valid_dist = {
    "alpha": alpha,
    "anglit": anglit,
    "arcsine": arcsine,
    "argus": argus,
    "beta": beta,
    "betaprime": betaprime,
    "bradford": bradford,
    "burr": burr,
    "burr12": burr12,
    "cauchy": cauchy,
    "chi": chi,
    "chi2": chi2,
    "cosine": cosine,
    "crystalball": crystalball,
    "dgamma": dgamma,
    "dweibull": dweibull,
    "erlang": erlang,
    "expon": expon,
    "exponnorm": exponnorm,
    "exponweib": exponweib,
    "exponpow": exponpow,
    "f": f,
    "fatiguelife": fatiguelife,
    "fisk": fisk,
    "foldcauchy": foldcauchy,
    "foldnorm": foldnorm,
    # "frechet_r": frechet_r,
    # "frechet_l": frechet_l,
    "genlogistic": genlogistic,
    "gennorm": gennorm,
    "genpareto": genpareto,
    "genexpon": genexpon,
    "genextreme": genextreme,
    "gausshyper": gausshyper,
    "gamma": gamma,
    "gengamma": gengamma,
    "genhalflogistic": genhalflogistic,
    # "geninvgauss": geninvgauss,
    "gilbrat": gilbrat,
    "gompertz": gompertz,
    "gumbel_r": gumbel_r,
    "gumbel_l": gumbel_l,
    "halfcauchy": halfcauchy,
    "halflogistic": halflogistic,
    "halfnorm": halfnorm,
    "halfgennorm": halfgennorm,
    "hypsecant": hypsecant,
    "invgamma": invgamma,
    "invgauss": invgauss,
    "invweibull": invweibull,
    "johnsonsb": johnsonsb,
    "johnsonsu": johnsonsu,
    "kappa4": kappa4,
    "kappa3": kappa3,
    "ksone": ksone,
    "kstwobign": kstwobign,
    "laplace": laplace,
    "levy": levy,
    "levy_l": levy_l,
    "levy_stable": levy_stable,
    "logistic": logistic,
    "loggamma": loggamma,
    "loglaplace": loglaplace,
    "lognorm": lognorm,
    # "loguniform": loguniform,
    "lomax": lomax,
    "maxwell": maxwell,
    "mielke": mielke,
    "moyal": moyal,
    "nakagami": nakagami,
    "ncx2": ncx2,
    "ncf": ncf,
    "nct": nct,
    "norm": norm,
    "norminvgauss": norminvgauss,
    "pareto": pareto,
    "pearson3": pearson3,
    "powerlaw": powerlaw,
    "powerlognorm": powerlognorm,
    "powernorm": powernorm,
    "rdist": rdist,
    "rayleigh": rayleigh,
    "rice": rice,
    "recipinvgauss": recipinvgauss,
    "skewnorm": skewnorm,
    "t": t,
    "trapz": trapz,
    "triang": triang,
    "truncexpon": truncexpon,
    "truncnorm": truncnorm,
    "tukeylambda": tukeylambda,
    "uniform": uniform,
    "vonmises": vonmises,
    "vonmises_line": vonmises_line,
    "wald": wald,
    "weibull_min": weibull_min,
    "weibull_max": weibull_max,
    "wrapcauchy": wrapcauchy,
}

param_dist = {
    "alpha": ["a", "loc", "scale"],
    "anglit": ["loc", "scale"],
    "arcsine": ["loc", "scale"],
    "argus": ["chi", "loc", "scale"],
    "beta": ["a", "b", "loc", "scale"],
    "betaprime": ["a", "b", "loc", "scale"],
    "bradford": ["c", "loc", "scale"],
    "burr": ["c", "d", "loc", "scale"],
    "burr12": ["c", "d", "loc", "scale"],
    "cauchy": ["loc", "scale"],
    "chi": ["df", "loc", "scale"],
    "chi2": ["df", "loc", "scale"],
    "cosine": ["loc", "scale"],
    "crystalball": ["beta", "m", "loc", "scale"],
    "dgamma": ["a", "loc", "scale"],
    "dweibull": ["c", "loc", "scale"],
    "erlang": ["loc", "scale"],
    "expon": ["loc", "scale"],
    "exponnorm": ["K", "loc", "scale"],
    "exponweib": ["a", "c", "loc", "scale"],
    "exponpow": ["b", "loc", "scale"],
    "f": ["dfn", "dfd", "loc", "scale"],
    "fatiguelife": ["c", "loc", "scale"],
    "fisk": ["c", "loc", "scale"],
    "foldcauchy": ["c", "loc", "scale"],
    "foldnorm": ["c", "loc", "scale"],
    "frechet_r": ["c", "loc", "scale"],
    "frechet_l": ["c", "loc", "scale"],
    "genlogistic": ["c", "loc", "scale"],
    "gennorm": ["beta", "loc", "scale"],
    "genpareto": ["c", "loc", "scale"],
    "genexpon": ["a", "b", "c", "loc", "scale"],
    "genextreme": ["c", "loc", "scale"],
    "gausshyper": ["a", "b", "c", "z", "loc", "scale"],
    "gamma": ["a", "loc", "scale"],
    "gengamma": ["a", "c", "loc", "scale"],
    "genhalflogistic": ["c", "loc", "scale"],
    "geninvgauss": ["p", "b", "loc", "scale"],
    "gilbrat": ["loc", "scale"],
    "gompertz": ["c", "loc", "scale"],
    "gumbel_r": ["loc", "scale"],
    "gumbel_l": ["loc", "scale"],
    "halfcauchy": ["loc", "scale"],
    "halflogistic": ["loc", "scale"],
    "halfnorm": ["loc", "scale"],
    "halfgennorm": ["beta", "loc", "scale"],
    "hypsecant": ["loc", "scale"],
    "invgamma": ["a", "loc", "scale"],
    "invgauss": ["mu", "loc", "scale"],
    "invweibull": ["c", "loc", "scale"],
    "johnsonsb": ["a", "b", "loc", "scale"],
    "johnsonsu": ["a", "b", "loc", "scale"],
    "kappa4": ["h", "k", "loc", "scale"],
    "kappa3": ["a", "loc", "scale"],
    "ksone": ["n", "loc", "scale"],
    "kstwobign": ["loc", "scale"],
    "laplace": ["loc", "scale"],
    "levy": ["loc", "scale"],
    "levy_l": ["loc", "scale"],
    "levy_stable": ["alpha", "beta", "loc", "scale"],
    "logistic": ["loc", "scale"],
    "loggamma": ["c", "loc", "scale"],
    "loglaplace": ["c", "loc", "scale"],
    "lognorm": ["s", "loc", "scale"],
    "loguniform": ["a", "b", "loc", "scale"],
    "lomax": ["c", "loc", "scale"],
    "maxwell": ["loc", "scale"],
    "mielke": ["k", "s", "loc", "scale"],
    "moyal": ["loc", "scale"],
    "nakagami": ["nu", "loc", "scale"],
    "ncx2": ["df", "nc", "loc", "scale"],
    "ncf": ["dfn", "dfd", "nc", "loc", "scale"],
    "nct": ["df", "nc", "loc", "scale"],
    "norm": ["loc", "scale"],
    "norminvgauss": ["a", "b", "loc", "scale"],
    "pareto": ["b", "loc", "scale"],
    "pearson3": ["skew", "loc", "scale"],
    "powerlaw": ["a", "loc", "scale"],
    "powerlognorm": ["c", "s", "loc", "scale"],
    "powernorm": ["c", "loc", "scale"],
    "rdist": ["c", "loc", "scale"],
    "rayleigh": ["loc", "scale"],
    "rice": ["b", "loc", "scale"],
    "recipinvgauss": ["mu", "loc", "scale"],
    "semicircular": ["loc", "scale"],
    "skewnorm": ["a", "loc", "scale"],
    "t": ["df", "loc", "scale"],
    "trapz": ["c", "d", "loc", "scale"],
    "triang": ["c", "loc", "scale"],
    "truncexpon": ["b", "loc", "scale"],
    "truncnorm": ["a", "b", "loc", "scale"],
    "tukeylambda": ["lam", "loc", "scale"],
    "uniform": ["loc", "scale"],
    "vonmises": ["kappa", "loc", "scale"],
    "vonmises_line": ["kappa", "loc", "scale"],
    "wald": ["loc", "scale"],
    "weibull_min": ["c", "loc", "scale"],
    "weibull_max": ["c", "loc", "scale"],
    "wrapcauchy": ["c", "loc", "scale"],
}

## Marginal classes
##################################################
# Marginal parent class
class Marginal(ABC):
    """Parent class for marginal distributions
    """

    def __init__(self, sign=0):
        self.sign = sign

    @abstractmethod
    def copy(self):
        pass

    ## Fitting function
    @abstractmethod
    def fit(self, data):
        pass

    ## Likelihood function
    @abstractmethod
    def l(self, x):
        pass

    ## Cumulative density function
    @abstractmethod
    def p(self, x):
        pass

    ## Quantile function
    @abstractmethod
    def q(self, p):
        pass

    ## Summary
    @abstractmethod
    def summary(self):
        pass


## Named marginal class
class MarginalNamed(Marginal):
    """Marginal using a named distribution from gr.valid_dist"""

    def __init__(self, d_name=None, d_param=None, **kw):
        super().__init__(**kw)

        self.d_name = d_name
        self.d_param = d_param

    def copy(self):
        new_marginal = MarginalNamed(
            sign=self.sign, d_name=self.d_name, d_param=copy.deepcopy(self.d_param)
        )

        return new_marginal

    ## Fitting function
    def fit(self, data):
        param = valid_dist[self.d_name].fit(data)
        self.d_param = dict(zip(param_dist[dist], param))

    ## Likelihood function
    @make_symbolic
    def l(self, x):
        return valid_dist[self.d_name].pdf(x, **self.d_param)

    ## Cumulative density function
    @make_symbolic
    def p(self, x):
        return valid_dist[self.d_name].cdf(x, **self.d_param)

    ## Quantile function
    @make_symbolic
    def q(self, p):
        return valid_dist[self.d_name].ppf(p, **self.d_param)

    ## Summary
    def summary(self):
        return "({0:+}) {1:}, {2:}".format(self.sign, self.d_name, self.d_param)


## Gaussian KDE marginal class
class MarginalGKDE(Marginal):
    """Marginal using scipy.stats.gaussian_kde"""

    def __init__(self, kde, atol=1e-6, **kw):
        super().__init__(**kw)

        self.kde = kde
        self.atol = atol
        self._set_bracket()

    def copy(self):
        new_marginal = MarginalGKDE(kde=copy.deepcopy(self.kde), atol=self.atol,)

        return new_marginal

    def _set_bracket(self):
        ## Calibrate the quantile brackets based on desired accuracy
        bracket = [npmin(self.kde.dataset), npmax(self.kde.dataset)]

        sol_lo = root_scalar(
            lambda x: self.atol - self.p(x),
            x0=bracket[0],
            x1=bracket[1],
            method="secant",
        )
        sol_hi = root_scalar(
            lambda x: 1 - self.atol - self.p(x),
            x0=bracket[0],
            x1=bracket[1],
            method="secant",
        )

        self.bracket = [sol_lo.root, sol_hi.root]

    ## Fitting function
    def fit(self, data):
        self.kde = gaussian_kde(data)
        self._set_bracket()

    ## Likelihood function
    @make_symbolic
    def l(self, x):
        return self.kde.pdf(x)

    ## Cumulative density function
    @make_symbolic
    def p(self, x):
        try:
            return array([self.kde.integrate_box_1d(-Inf, v) for v in x])
        except TypeError:
            return self.kde.integrate_box_1d(-Inf, x)

    ## Quantile function
    @make_symbolic
    def q(self, p):
        p_bnd = self.p(self.bracket)

        ## Create scalar solver
        def qscalar(val):
            if val <= p_bnd[0]:
                return self.bracket[0]
            if val >= p_bnd[1]:
                return self.bracket[1]
            sol = root_scalar(
                lambda x: val - self.p(x),
                bracket=self.bracket,
                method="bisect",
                xtol=self.atol,
            )
            return sol.root

        ## Iterate over all given values
        try:
            res = zeros(len(p))
            for i in range(len(p)):
                res[i] = qscalar(p[i])
        except TypeError:
            res = qscalar(p)

        return res

    ## Summary
    def summary(self):
        p_bnd = self.p(self.bracket)

        return "({0:+}) gaussian KDE, n={1:2.1f}/{2:}, f={3:2.1e}, ".format(
            self.sign, self.kde.neff, self.kde.dataset.shape[1], self.kde.factor
        ) + "b=[{0:2.1e}, {1:2.1e}], a={2:1.0e}".format(
            self.bracket[0], self.bracket[1], self.atol
        )

## Marginal functions
##################################################
def marg_mom(
        dist,
        mean=None,
        sd=None,
        cov=None,
        var=None,
        skew=None,
        kurt=None,
        kurt_excess=None,
        sign=0,
        dict_x0=None,
):
    r"""Fit scipy.stats continuous distribution via moments

    Fit a continuous distribution using the method of moments. Select a
    distribution shape and provide numerical values for a convenient set of
    common moments.

    This routine uses a vector-output root finding routine to match the moments.
    You may set an optional initial guess for the distribution parameters using
    the dict_x0 argument.

    Args:
        dist (str): Name of distribution to fit

    Kwargs:
        mean (float): Mean of distribution
        sd (float): Standard deviation of distribution
        cov (float): Coefficient of Variation of distribution (sd / mean)
        var (float): Variance of distribution; only one of `sd` and `var` can be provided.
        skew (float): Skewness of distribution
        kurt (float): Kurtosis of distribution
        kurt_excess (float): Excess kurtosis of distribution; kurt_excess = kurt - 3.
            Only one of `kurt` and `kurt_excess` can be provided.

        sign (-1, 0, +1): Sign
        dict_x0 (dict): Dictionary of initial parameter guesses

    Returns:
        gr.MarginalNamed: Distribution

    Examples:
        >>> import grama as gr
        >>> ## Fit a normal distribution
        >>> mg_norm = gr.marg_mom("norm", mean=0, sd=1)
        >>> ## Fit a lognormal distribution
        >>> mg_lognorm = gr.marg_mom("lognorm", mean=1, sd=1, skew=1)
        >>> ## Fit a lognormal, controlling kurtosis instead
        >>> mg_lognorm = gr.marg_mom("lognorm", mean=1, sd=1, kurt=1)
        >>>
        >>> ## Not all moment combinations are feasible; this will fail
        >>> gr.marg_mom("beta", mean=1, sd=1, skew=0, kurt=4)
        >>> ## Skewness and kurtosis are related for the beta distribution;
        >>> ## a different combination is feasible
        >>> gr.marg_mom("beta", mean=1, sd=1, skew=0, kurt=2)

    """
    ## Number of distribution parameters
    n_param = len(param_dist[dist])
    if n_param > 4:
        raise NotImplementedError(
            "marg_nom does not yet handle distributions with more than 4 parameters"
        )

    ## Check invariants
    if mean is None:
        raise ValueError("Must provide `mean` argument.")
    if (sd is None) and (var is None) and (cov is None):
        raise ValueError(
            "One of `sd`, `cov`, or `var` must be provided."
        )
    if sum([(not sd is None), (not var is None), (not cov is None)]) > 1:
        raise ValueError(
            "Only one of `sd`, `cov`, and `var` may be provided."
        )
    if (not kurt is None) and (not kurt_excess is None):
        raise ValueError(
            "Only one of `kurt` and `kurt_excess` may be provided."
        )

    ## Process arguments
    # Transform to "standard" moments
    if (not sd is None):
        var = sd**2
    if (not cov is None):
        var = (mean * cov)**2
    if (not kurt is None):
        kurt_excess = kurt - 3

    # Build up target moments
    s = "mv"
    m_target = array([mean, var])

    if (not skew is None):
        s = s + "s"
        m_target = concatenate((m_target, array([skew])))
    if (not kurt is None):
        s = s + "k"
        m_target = concatenate((m_target, array([kurt_excess])))
    n_provided = len(s)

    if n_provided < n_param:
        raise ValueError(
            "Insufficient moments provided; you must provide {} more moment(s).".format(
                n_param - n_provided
            )
        )
    if n_provided > n_param:
        raise ValueError(
            "Overdetermined; you must provide {} fewer moment(s).".format(
                n_provided - n_param
            )
        )

    ## Generate helper function for optimization
    def _obj(v):
        kw = dict(zip(param_dist[dist], v))
        return array(valid_dist[dist](**kw).stats(s)) - m_target

    ## Generate initial guess
    if dict_x0 is None:
        # Manually coded initial guesses
        dict_x0 = dict(
            loc=mean,
            scale=sqrt(var),
            a=1,
            b=1,
            s=1,
            df=10,
            #K=None,
            #chi=None,
        )
    # Repackage for optimizer
    x0 = array([
        dict_x0[key] for key in param_dist[dist]
    ])

    ## Run multidimensional root finding
    res = root(
        _obj,
        x0
    )

    ## Check for failed optimization
    if res.success is False:
        raise RuntimeError(
            "Moment matching failed; initial guess may be poor, or requested "
            "moments may be infeasible. Try setting `dict_x0`. " +
            "Printing optimization results for debugging:\n\n{}".format(res)
        )

    ## Repackage and return
    param = dict(zip(param_dist[dist], res.x))
    return MarginalNamed(sign=sign, d_name=dist, d_param=param)

## Fit a named scipy.stats distribution
def marg_named(data, dist, name=True, sign=None):
    r"""Fit scipy.stats continuous distirbution

    Fits a named scipy.stats continuous distribution. Intended to be used to
    define a marginal distribution from data.

    Args:
        data (iterable): Data for fit
        dist (str): Distribution to fit
        name (bool): Include distribution name?
        sign (bool): Include sign? (Optional)

    Returns:
        gr.MarginalNamed: Distribution

    Examples:

        >>> import grama as gr
        >>> from grama.data import df_stang
        >>>     gr.cp_marginals(
        >>>         E=gr.marg_named(df_stang.E, "norm"),
        >>>         mu=gr.marg_named(df_stang.mu, "beta")
        >>>     )
        >>> md.printpretty()

    """
    ## Catch case where user provides entire DataFrame
    if isinstance(data, DataFrame):
        raise ValueError("`data` argument must be a single column; try data.var")

    ## Fit the distribution
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        param = valid_dist[dist].fit(data)

    param = dict(zip(param_dist[dist], param))

    if sign is not None:
        if not (sign in [-1, 0, +1]):
            raise ValueError("Invalid `sign`")
    else:
        sign = 0

    return MarginalNamed(sign=sign, d_name=dist, d_param=param)


## Fit a gaussian kernel density estimate (KDE) to data
def marg_gkde(data, sign=None):
    r"""Fit a gaussian KDE to data

    Fits a gaussian kernel density estimate (KDE) to data.

    Args:
        data (iterable): Data for fit
        sign (bool): Include sign? (Optional)

    Returns:
        gr.MarginalGKDE: Marginal distribution

    Examples:

        >>> import grama as gr
        >>> from grama.data import df_stang
        >>> md = gr.Model("Marginal Example") >> \
        >>>     gr.cp_marginals(
        >>>         E=gr.marg_gkde(df_stang.E),
        >>>         mu=gr.marg_gkde(df_stang.mu)
        >>>     )
        >>> md.printpretty()

    """
    ## Catch case where user provides entire DataFrame
    if isinstance(data, DataFrame):
        raise ValueError("`data` argument must be a single column; try data.var")

    kde = gaussian_kde(data)
    if sign is not None:
        if not (sign in [-1, 0, +1]):
            raise ValueError("Invalid `sign`")
    else:
        sign = 0

    return MarginalGKDE(kde, sign=sign)



__all__ = ["make_composite_plate_tension"]

##
#
# References:
# C.T. Sun "Mechanics of Aircraft Structures" (1998)

import itertools
from grama import CopulaIndependence, Density, Domain, Function, \
    MarginalNamed, Model
from numpy import pi, sqrt, cos, sin, array, dot, zeros, reshape, Inf, tile
from numpy import abs as npabs
from numpy import sum as npsum
from numpy.linalg import solve, inv


## Composite properties
##################################################
# Inspired by Arevo data
E1_M   = 114e9
E2_M   =   7e9
G12_M  =   4e9
nu12_M = 0.45

E1_CV    = 0.02
E2_CV    = 0.08
G12_CV   = 0.1
nu12_SIG = nu12_M * 0.08 # By 8% COV

T_NOM    = 1e-3            # Nominal thickness
T_PM     = T_NOM * 0.01    # +/- ply thickness
THETA_PM = 3 * pi / 180 # +/- angle tolerance

SIG_11_T_M = 1.4e9 # Tensile strength
SIG_11_C_M = 0.5e9 # Compressive strength
SIG_12_M_M =  62e6 # Shear strength

SIG_11_T_CV = 0.06
SIG_11_C_CV = 0.06
SIG_12_M_CV = 0.07

SIG_22_T_M = 1.4e6 # Tensile strength
SIG_22_C_M = 0.5e6 # Compressive strength

SIG_22_T_CV = 0.06
SIG_22_C_CV = 0.06

Nx_M = 1.2e6                    # Nominal load conditions [N/m]
Nx_SIG = Nx_M * sqrt(0.01)

## Laminate Composite Analysis
##################################################
# Construct stress transformation matrix
def make_Ts(theta):
    """
    Usage
        Ts    = make_Ts(theta)
    Arguments
        theta = angle of rotation; in radians
    Returns
        Ts    = stress transformation matrix
    """
    C = cos(theta)
    S = sin(theta)

    return array([
        [  C**2,  S**2,   2 * S * C],
        [  S**2,  C**2,  -2 * S * C],
        [-S * C, S * C, C**2 - S**2]
    ])

# Construct strain transformation matrix
def make_Te(theta):
    """
    Usage
        Te    = make_Te(theta)
    Arguments
        theta = angle of rotation; in radians
    Returns
        Te    = strain transformation matrix
    """
    C = cos(theta)
    S = sin(theta)

    return array([
        [      C**2,      S**2,       S * C],
        [      S**2,      C**2,      -S * C],
        [-2 * S * C, 2 * S * C, C**2 - S**2]
    ])

# Construct stiffness matrix in lamina coordinate system
def make_Q(param):
    """
    Usage
        Q     = make_Q(param)
    Arguments
        param = array of parameters
              = [E1, E2, nu12, G12]
    Returns
        Q     = stiffness matrix in lamina coordinate system
    """
    E1, E2, nu12, G12 = param

    nu21 = nu12 * E2 / E1

    Q11 = E1 / (1 - nu12 * nu21)
    Q12 = nu12 * E2 / (1 - nu12 * nu21)
    Q21 = Q12
    Q22 = E2 / (1 - nu12 * nu21)
    Q66 = G12

    return array([
        [ Q11, Q12,   0],
        [ Q21, Q22,   0],
        [   0,   0, Q66]
    ])

# Construct stiffness matrix in global coordinate system
def make_Qb(param, theta):
    """
    Usage
        Qb    = make_Qb(param)
    Arguments
        param = array of parameters
              = [E1, E2, nu12, G12]
        theta = lamina angle; in radians
    Returns
        Qb    = stiffness matrix in global coordinate system
    """
    Ts_inv = make_Ts(-theta)
    Te     = make_Te(+theta)
    Q      = make_Q(param)

    return dot(Ts_inv, dot(Q, Te))

# Construct extensional stiffness matrix
def make_A(Param, Theta, T):
    """
    Usage
        A     = make_A(Param, Theta, T)
    Arguments
        Param = array of array of parameters
              = [[E1, E2, nu12, G12]_1,
                          . . .
                 [E1, E2, nu12, G12]_k]
        Theta = array of lamina angles
              = [theta_1, ..., theta_k]
        T     = array of thicknesses
              = [t_1, ..., t_k]
    Returns
        A     = extensional stiffness matrix

    @pre len(Param) == len(Theta)
    @pre len(Param) == len(T)
    """
    n_k = len(Param)
    Qb_all = zeros((3, 3, n_k))

    for ind in range(n_k):
        Qb_all[:, :, ind] = make_Qb(Param[ind], Theta[ind]) * T[ind]

    return npsum(Qb_all, axis = 2)

# Compute stresses under uniaxial tension
def uniaxial_stresses(Param, Theta, T):
    """
    Return the lamina stresses for a unit uniaxial tension along the fiber direction

    Usage
        Stresses = uniaxial_stresse(Param, Theta, T)
    Arguments
        Param = array of array of parameters
              = [[E1, E2, nu12, G12]_1,
                          . . .
                 [E1, E2, nu12, G12]_k]
        Theta = array of lamina angles
              = [theta_1, ..., theta_k]
        T     = array of thicknesses
              = [t_1, ..., t_k]
    Returns
        Stresses = state of stress in each lamina
                 = [[\sigma_11, \sigma_22, \sigma_12]_1
                                 . . .

                 =  [\sigma_11, \sigma_22, \sigma_12]_k]
    """
    ## Solve for state of strain with uniaxial loading
    Ab = make_A(Param, Theta, T)
    strains = solve(Ab, array([1, 0, 0]))

    ## Solve for each lamina state of strain in local coordinates
    n_k      = len(Param)
    Stresses = zeros((n_k, 3))
    for ind in range(n_k):
        ## Solve in global coordinates
        stress_global = dot(make_Qb(Param[ind], Theta[ind]), strains)
        ## Convert to lamina coordinates
        Stresses[ind] = dot(make_Ts(Theta[ind]), stress_global)

    return Stresses

# Uniaxial tension limit state functions
def uniaxial_stress_limit(X):
    """
    Evaluate stress limit states for uniaxial tension. Dimensionality of problem
    is inferred from size of X.

    Usage
        g_stress = uniaxial_stress_limit(X)
    Arguments
        X = array of composite laminate properties and loading
          = [E1, E2, nu12, G12, theta, t, ...
              sig_11_t, sig_22_t, sig_11_c, sig_22_c, sig_12_s, # for i = 1
                                  .  .  .
             E1, E2, nu12, G12, theta, t, ...
              sig_11_t, sig_22_t, sig_11_c, sig_22_c, sig_12_s, # for i = k
             Nx]
    Returns
        g_stress = array of limit state values
                 = [g_11_t_1 g_22_t_1 g_11_c_1 g_22_c_1 g_s_12_1,
                            .  .  .
                    g_11_t_k g_22_t_k g_11_c_k g_22_c_k g_s_12_k]
    @pre ((len(X) - 1) % 11) == 0
    """
    ## Pre-process inputs
    k = int((len(X) - 1) / 11)
    Y = reshape(array(X[:-1]), (k, 11))

    ## Unpack inputs
    Nx        = X[-1]
    Param     = Y[:, 0:4]  # [E1, E2, nu12, G12]
    Theta     = Y[:, 4]
    T         = Y[:, 5]
    Sigma_max = Y[:, 6:11] # [sig_11_t, sig_22_t, sig_11_c, sig_22_c, sig_12_s]

    ## Evaluate stress [\sigma_11, \sigma_22, \sigma_12]_i
    Stresses = Nx * uniaxial_stresses(Param, Theta, T)

    ## Construct limit state
    g_limit = zeros((k, 5))
    g_limit[:, (0,1)] = +1 - Stresses[:, (0,1)] / Sigma_max[:, (0,1)]
    g_limit[:, (2,3)] = +1 + Stresses[:, (0,1)] / Sigma_max[:, (2,3)]
    g_limit[:, 4]     = +1 - npabs(Stresses[:, 2]) / Sigma_max[:, 4]

    return g_limit.flatten()

## Random variable model
##################################################
def make_names(Theta_nom):
    k = len(Theta_nom)
    vars_list = list(itertools.chain.from_iterable([
       ["E1_{}".format(i),
        "E2_{}".format(i),
        "nu12_{}".format(i),
        "G12_{}".format(i),
        "theta_{}".format(i),
        "t_{}".format(i),
        "sigma_11_t_{}".format(i),
        "sigma_22_t_{}".format(i),
        "sigma_11_c_{}".format(i),
        "sigma_22_c_{}".format(i),
        "sigma_12_s_{}".format(i)] for i in range(k)
    ])) + ["Nx"]

    return vars_list

# Helper function to create domain for given ply
def make_domain(Theta_nom, T_nom= T_NOM):
    """
    Helper function to construct domain object for composite panel

    Usage
        domain = make_domain(Theta_nom)
        domain = make_domain(Theta_nom, T_nom= T_nom)
    Arguments
        Theta_nom = nominal lamina angles; determines dimensionality of problem
                  = [theta_1, ..., theta_k]
    Keyword Arguments
        t_nom     = nominal ply thicknesses
                  = [t_1, ..., t_k] OR
                  = t_nom
    Returns
        domain = grama domain object
    """
    k = len(Theta_nom)
    bounds_list = list(itertools.chain.from_iterable([
       [("E1_{}".format(i), [0, +Inf]),
        ("E2_{}".format(i), [0, +Inf]),
        ("nu12_{}".format(i), [-Inf, +Inf]),
        ("G12_{}".format(i), [0, +Inf]),
        ("theta_{}".format(i), [-pi/2, +pi/2]),
        ("t_{}".format(i), [-Inf, +Inf]),
        ("sigma_11_t_{}".format(i), [0, +Inf]),
        ("sigma_22_t_{}".format(i), [0, +Inf]),
        ("sigma_11_c_{}".format(i), [0, +Inf]),
        ("sigma_22_c_{}".format(i), [0, +Inf]),
        ("sigma_12_s_{}".format(i), [0, +Inf])] for i in range(k)
    ])) + [("Nx", [-Inf, +Inf])]
    bounds = dict(bounds_list)

    return Domain(bounds=bounds)

# Helper function to create density for given ply
def make_density(Theta_nom, T_nom=T_NOM):
    """
    Helper function to construct density object for composite panel

    Usage
        density = make_density(Theta_nom)
        density = make_density(Theta_nom, t_nom= t_nom)
    Arguments
        Theta_nom = nominal lamina angles; determines dimensionality of problem
                      = [theta_1, ..., theta_k]
    Keyword Arguments
        t_nom     = nominal ply thicknesses
                      = [t_1, ..., t_k] OR
                      = t_nom
    Returns
        density = grama density object
    """
    ## Setup
    k = len(Theta_nom)
    if not isinstance(T_nom, list):
        T_nom = [T_nom] * k

    ## Create variables for each ply
    marginals = {}
    for i in range(k):
        marginals["E1_{}".format(i)] = MarginalNamed(
            sign=0,
            d_name="lognorm",
            d_param={"loc": 1, "s": E1_CV, "scale": E1_M}
        )
        marginals["E2_{}".format(i)] = MarginalNamed(
            sign=0,
            d_name="lognorm",
            d_param={"loc": 1, "s": E2_CV, "scale": E2_M}
        )
        marginals["nu12_{}".format(i)] = MarginalNamed(
            sign=0,
            d_name="norm",
            d_param={"loc": nu12_M, "scale": nu12_SIG}
        )
        marginals["G12_{}".format(i)] = MarginalNamed(
            sign=0,
            d_name="lognorm",
            d_param={"loc": 1, "s": G12_CV, "scale": G12_M}
        )
        marginals["theta_{}".format(i)] = MarginalNamed(
            sign=0,
            d_name="uniform",
            d_param={"loc": Theta_nom[i] - THETA_PM, "scale": 2 * THETA_PM}
        )
        marginals["t_{}".format(i)] = MarginalNamed(
            sign=0,
            d_name="uniform",
            d_param={"loc": T_nom[i] - T_PM, "scale": 2 * T_PM}
           )
        marginals["sigma_11_t_{}".format(i)] = MarginalNamed(
            sign=-1,
            d_name="lognorm",
            d_param={"loc": 1, "s": SIG_11_T_CV, "scale": SIG_11_T_M}
           )
        marginals["sigma_22_t_{}".format(i)] = MarginalNamed(
            sign=-1,
            d_name="lognorm",
            d_param={"loc": 1, "s": SIG_22_T_CV, "scale": SIG_22_T_M}
           )
        marginals["sigma_11_c_{}".format(i)] = MarginalNamed(
            sign=-1,
            d_name="lognorm",
            d_param={"loc": 1, "s": SIG_11_C_CV, "scale": SIG_11_C_M}
           )
        marginals["sigma_22_c_{}".format(i)] = MarginalNamed(
            sign=-1,
            d_name="lognorm",
            d_param={"loc": 1, "s": SIG_22_C_CV, "scale": SIG_22_C_M}
           )
        marginals["sigma_12_s_{}".format(i)] = MarginalNamed(
            sign=-1,
            d_name="lognorm",
            d_param={"loc": 1, "s": SIG_12_M_CV, "scale": SIG_12_M_M}
        )

    marginals["Nx"] = MarginalNamed(
        sign=+1,
        d_name="norm",
        d_param={"loc": Nx_M, "scale": Nx_SIG}
    )
    var_rand = list(marginals.keys())

    return Density(
        marginals=marginals,
        copula=CopulaIndependence(var_rand)
    )

## Model class
##################################################
class make_composite_plate_tension(Model):
    def __init__(self, Theta_nom, T_nom=T_NOM):
        k = len(Theta_nom)
        deg_int = [int(theta / pi * 180) for theta in Theta_nom]
        def mapSign(x):
            if x < 0:
                return "m" + str(npabs(x))
            if x > 0:
                return "p" + str(x)
            return str(x)
        deg_str = map(mapSign, deg_int)
        name =  "Composite Plate in Tension " + "-".join(deg_str)

        super().__init__(
            name=name,
            functions=[
                Function(
                    lambda X: uniaxial_stress_limit(X),
                    make_names(Theta_nom),
                    list(itertools.chain.from_iterable([
                        ["g_11_tension_{}".format(i),
                         "g_22_tension_{}".format(i),
                         "g_11_compression_{}".format(i),
                         "g_22_compression_{}".format(i),
                         "g_12_shear_{}".format(i)] for i in range(k)
                    ])),
                    "limit states",
                    0
                )
            ],
            domain=make_domain(Theta_nom, T_nom=T_nom),
            density=make_density(Theta_nom, T_nom=T_nom)
        )

## Verification
##################################################
if __name__ == "__main__":
    ## Verify the stiffness matrix
    E1   = 140e9
    E2   = 10e9
    G12  = 6.9e9
    nu12 = 0.3

    param = [E1, E2, nu12, G12]
    theta = 45. * pi / 180

    Qb = make_Qb(param, theta)
    Sb = inv(Qb)

    ## Example 8.4
    E1   = 180e9    # Pa
    E2   =  10e9    # Pa
    G12  =   7e9    # Pa
    nu12 = 0.3      # [-]
    t    = 0.127e-3 # m

    Theta = array([+45, -45, 0, 90, 90, 0, -45, +45]) * pi / 180
    Param = tile([E1, E2, nu12, G12], reps = (len(Theta), 1))
    T     = array([t] * len(Theta))

    A = make_A(Param, Theta, T)

    # Check values; close to Sun (1998)
    print(A / t)

    # Uniaxial (unit) tension
    strain_uniaxial = linalg.solve(A, array([1, 0, 0]))
    stress_uniaxial = uniaxial_stresses(Param, Theta, T)

    ## Test model
    model = make_composite_plate_tension(
        Theta_nom = [0]
    )



__all__ = [
    "tran_reweight",
    "tf_reweight",
]

from grama import add_pipe, pipe, custom_formatwarning
from pandas import concat, DataFrame
from toolz import curry

## Reweight using likelihood ratio
# --------------------------------------------------
@curry
def tran_reweight(
        df_base,
        md_base,
        md_new,
        var_weight="weight",
        append=True,
):
    r"""Reweight a sample using likelihood ratio

    Reweight is a tool to facilitate "What If?" Monte Carlo simulation;
    specifically, to make testing a models with the same function(s) but
    different distributions more computationally efficient.

    This tool automates calulation of the *likelihood ratio* between the
    distributions of two given models. Using the resulting weights to scale
    (elementwise multiply) output values and compute summaries is called
    *importance sampling*, enabling "What If?" testing. Use of this tool enables
    one to generate a single Monte Carlo sample, rather than multiple samples
    for each "What If?" scenario (avoiding extraneous function evaluations).

    Let `y` be a generic output of the scenario. The importance sampling
    procedure is as follows:

    1. Create a base scenario represented by `md_base`, and a desired number
       of alternative "What If?" scenarios represented my other models.
    2. Use `gr.eval_monte_carlo` to generate a single sample `df_base` of size `n`
       using the base scenario `md_base`. Compute statistics of interest on
       the output `y` for this base scenario.
    3. For each alternative scenario `md_new`, use `gr.tran_reweight` to
       generate weights `weight`, and use the tool `n_e = gr.neff_is(DF.weight)`
       to compute the effective sample size. If `n_e << n`, then importance
       sampling is unlikely to give an accurate estimate for this scenario.
    4. For each alternative scenario `md_new`, use the relevant weights
       `weight` to scale the output value `y_new = y * weight`, and compute
       statistics of interest of the weighted output values.

    Args:
        df_base (DataFrame): Monte Carlo results from `md_base`.
        md_base (Model): Model used to generate `df_base`.
        md_new (Model): Model defining a new "What If?" scenario.
        var_weight (string): Name to give new column of weights.
        append (bool): Append results to original DataFrame?

    Returns:
        DataFrame: Original df_base with added column of weights.

    Notes:
        - The base scenario `md_base` should have fatter tails than any of
          the scenarios considered as `df_new`. See Owen (2013) Chapter 9
          for more details.

    References:
        A.B. Owen, "Monte Carlo theory, methods and examples" (2013)

    Examples:

        >>> import grama as gr
        >>> from grama.models import make_cantilever_beam
        >>> DF = gr.Intention()
        >>>
        >>> md_base = make_cantilever_beam()
        >>> md_new = (
        >>>     md_base
        >>>     >> gr.cp_marginals(
        >>>         H=dict(dist="norm", loc=500.0, scale=50.0),
        >>>     )
        >>> )
        >>>
        >>> ## Assess safety via simple Monte Carlo
        >>> df_base = gr.eval_monte_carlo(md_base, df_det="nom", n=1e3)
        >>> print(
        >>>     df_base
        >>>     >> gr.tf_summarize(
        >>>         pof_stress=gr.mean(DF.g_stress <= 0),
        >>>         pof_disp=gr.mean(DF.g_disp <= 0),
        >>>     )
        >>> )
        >>>
        >>> ## Re-use samples to test another scenario
        >>> print(
        >>>     df_base
        >>>     >> gr.tf_reweight(md_base=md_base, md_new=md_new)
        >>>     >> gr.tf_summarize(
        >>>         pof_stress=gr.mean((DF.g_stress <= 0) * DF.weight),
        >>>         pof_disp=gr.mean((DF.g_disp <= 0) * DF.weight),
        >>>         n_eff=gr.neff_is(DF.weight),
        >>>     )
        >>> )
        >>>
        >>> ## It is unsafe to study new scenarios with wider uncertainty than the base
        >>> ## scenario
        >>> md_poor = (
        >>>     md_base
        >>>     >> gr.cp_marginals(
        >>>         H=dict(dist="norm", loc=500.0, scale=400.0),
        >>>     )
        >>> )
        >>> ## Note the tiny effective size in this case
        >>> print(
        >>>     md_base
        >>>     >> gr.ev_monte_carlo(n=1e3, df_det="nom")
        >>>     >> gr.tf_reweight(md_base=md_base, md_new=md_poor)
        >>>     >> gr.tf_summarize(
        >>>         pof_stress=gr.mean((DF.g_stress <= 0) * DF.weight),
        >>>         pof_disp=gr.mean((DF.g_disp <= 0) * DF.weight),
        >>>         n_eff=gr.neff_is(DF.weight),
        >>>     )
        >>> )

    """
    ## Check invariants
    # Check that random inputs to md_base available in df_base
    var_diff = set(md_base.var_rand).difference(set(df_base.columns))
    if not (len(var_diff) == 0):
        raise ValueError(
            "Random variables in md_base missing from df_base:\n" +
            "Missing: {}".format(var_diff)
        )
    # Check that random inputs match between models
    var_base = set(md_base.var_rand)
    var_new = set(md_new.var_rand)
    if var_base != var_new:
        raise ValueError(
            "Random variables of md_base and md_var must match:\n" +
            "md_base is missing: {}\n".format(var_new.difference(var_base)) +
            "md_new is missing: {}".format(var_base.difference(var_new))
        )
    # Check that deterministic inputs match between models
    var_base = set(md_base.var_det)
    var_new = set(md_new.var_det)
    if var_base != var_new:
        raise ValueError(
            "Deterministic variables of md_base and md_var must match:\n" +
            "md_base is missing: {}\n".format(var_new.difference(var_base)) +
            "md_new is missing: {}".format(var_base.difference(var_new))
        )
    # Check that `weights` name does not collide
    if (var_weight in df_base.columns) and append:
        raise ValueError(
            "Weight name {} already in df_base.columns; ".format(var_weight) +
            "choose a new name."
        )

    ## Compute weight values
    # Use base model for importance distribution
    q = md_base.density.l(df_base)
    # Use new model for nominal distribution
    p = md_new.density.l(df_base)
    # Compute likelihood ratio
    w = p / q

    ## Return results
    df_res = DataFrame({var_weight : w})

    if append:
        df_res = concat(
            [df_base.reset_index(drop=True), df_res],
            axis=1,
        )

    return df_res

tf_reweight = add_pipe(tran_reweight)



__all__ = [
    "lead",
    "lag",
    "between",
    "dense_rank",
    "min_rank",
    "cumsum",
    "cummean",
    "cumsd",
    "cummax",
    "cummin",
    "cumprod",
    "cumany",
    "cumall",
    "percent_rank",
    "row_number",
]

from .base import make_symbolic


# ------------------------------------------------------------------------------
# Window functions
# ------------------------------------------------------------------------------


@make_symbolic
def lead(series, i=1):
    """
    Returns a series shifted forward by a value. `NaN` values will be filled
    in the end.

    Same as a call to `series.shift(i)`

    Args:
        series: column to shift forward.
        i (int): number of positions to shift forward.
    """

    shifted = series.shift(i * -1)
    return shifted


@make_symbolic
def lag(series, i=1):
    """
    Returns a series shifted backwards by a value. `NaN` values will be filled
    in the beginning.

    Same as a call to `series.shift(-i)`

    Args:
        series: column to shift backward.
        i (int): number of positions to shift backward.
    """

    shifted = series.shift(i)
    return shifted


@make_symbolic
def between(series, a, b, inclusive=False):
    """
    Returns a boolean series specifying whether rows of the input series
    are between values `a` and `b`.

    Args:
        series: column to compare, typically symbolic.
        a: value series must be greater than (or equal to if `inclusive=True`)
            for the output series to be `True` at that position.
        b: value series must be less than (or equal to if `inclusive=True`) for
            the output series to be `True` at that position.

    Kwargs:
        inclusive (bool): If `True`, comparison is done with `>=` and `<=`.
            If `False` (the default), comparison uses `>` and `<`.
    """

    if inclusive == True:
        met_condition = (series >= a) & (series <= b)
    elif inclusive == False:
        met_condition = (series > a) & (series < b)
    return met_condition


@make_symbolic
def dense_rank(series, ascending=True):
    """
    Equivalent to `series.rank(method='dense', ascending=ascending)`.

    Args:
        series: column to rank.

    Kwargs:
        ascending (bool): whether to rank in ascending order (default is `True`).
    """

    ranks = series.rank(method="dense", ascending=ascending)
    return ranks


@make_symbolic
def min_rank(series, ascending=True):
    """
    Equivalent to `series.rank(method='min', ascending=ascending)`.

    Args:
        series: column to rank.

    Kwargs:
        ascending (bool): whether to rank in ascending order (default is `True`).
    """

    ranks = series.rank(method="min", ascending=ascending)
    return ranks


@make_symbolic
def cumsum(series):
    """
    Calculates cumulative sum of values. Equivalent to `series.cumsum()`.

    Args:
        series: column to compute cumulative sum for.
    """

    sums = series.cumsum()
    return sums


@make_symbolic
def cummean(series):
    """
    Calculates cumulative mean of values. Equivalent to
    `series.expanding().mean()`.

    Args:
        series: column to compute cumulative mean for.
    """

    means = series.expanding().mean()
    return means


@make_symbolic
def cumsd(series):
    """
    Calculates cumulative standard deviation of values. Equivalent to
    `series.expanding().sd()`.

    Args:
        series: column to compute cumulative sd for.
    """

    sds = series.expanding(2).std()
    return sds


@make_symbolic
def cummax(series):
    """
    Calculates cumulative maximum of values. Equivalent to
    `series.expanding().max()`.

    Args:
        series: column to compute cumulative maximum for.
    """

    maxes = series.expanding().max()
    return maxes


@make_symbolic
def cummin(series):
    """
    Calculates cumulative minimum of values. Equivalent to
    `series.expanding().min()`.

    Args:
        series: column to compute cumulative minimum for.
    """

    mins = series.expanding().min()
    return mins


@make_symbolic
def cumprod(series):
    """
    Calculates cumulative product of values. Equivalent to
    `series.cumprod()`.

    Args:
        series: column to compute cumulative product for.
    """

    prods = series.cumprod()
    return prods


@make_symbolic
def cumany(series):
    """
    Calculates cumulative any of values. Equivalent to
    `series.expanding().apply(np.any).astype(bool)`.

    Args:
        series: column to compute cumulative any for.
    """

    anys = series.expanding().apply(np.any).astype(bool)
    return anys


@make_symbolic
def cumall(series):
    """
    Calculates cumulative all of values. Equivalent to
    `series.expanding().apply(np.all).astype(bool)`.

    Args:
        series: column to compute cumulative all for.
    """

    alls = series.expanding().apply(np.all).astype(bool)
    return alls


@make_symbolic
def percent_rank(series, ascending=True):
    if series.size == 1:
        return 0
    percents = (series.rank(method="min", ascending=ascending) - 1) / (series.size - 1)
    return percents


@make_symbolic
def row_number(series, ascending=True):
    """
    Returns row number based on column rank
    Equivalent to `series.rank(method='first', ascending=ascending)`.

    Args:
        series: column to rank.

    Kwargs:
        ascending (bool): whether to rank in ascending order (default is `True`).

    Usage:
    diamonds >> head() >> mutate(rn=row_number(X.x))

       carat      cut color clarity  depth  table  price     x     y     z   rn
    0   0.23    Ideal     E     SI2   61.5   55.0    326  3.95  3.98  2.43  2.0
    1   0.21  Premium     E     SI1   59.8   61.0    326  3.89  3.84  2.31  1.0
    2   0.23     Good     E     VS1   56.9   65.0    327  4.05  4.07  2.31  3.0
    3   0.29  Premium     I     VS2   62.4   58.0    334  4.20  4.23  2.63  4.0
    4   0.31     Good     J     SI2   63.3   58.0    335  4.34  4.35  2.75  5.0
    """

    series_rank = series.rank(method="first", ascending=ascending)
    return series_rank
